# workflow_engine/core/values/data.py
import asyncio
import json
import logging
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, ClassVar, Generic, Type, TypeVar

from pydantic import ConfigDict, create_model

from ...utils.immutable import ImmutableBaseModel
from .mapping import StringMapValue
from .value import Caster, Value, ValueType, get_origin_and_args

if TYPE_CHECKING:
    from ..context import Context
    from .schema import ValueSchema

logger = logging.getLogger(__name__)


class Data(ImmutableBaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(extra="forbid")

    def __init_subclass__(cls, **kwargs):
        """Ensure all fields in subclasses are Value types."""
        super().__init_subclass__(**kwargs)

        for field_name, field_info in cls.model_fields.items():
            if not issubclass(field_info.annotation, Value):  # type: ignore
                raise TypeError(
                    f"Field '{field_name}' in {cls.__name__} must be a Value type, got {field_info.annotation}"
                )

    def to_dict(self) -> Mapping[str, Value]:
        data: dict[str, Value] = {}
        for key in self.__class__.model_fields.keys():
            value = getattr(self, key)
            assert isinstance(value, Value)
            data[key] = value
        return data

    @classmethod
    def to_value_schema(cls) -> "ValueSchema":
        from .schema import validate_value_schema  # avoid circular import

        return validate_value_schema(cls.model_json_schema())


type DataMapping = Mapping[str, Value]


def dump_data_mapping(data: DataMapping) -> Mapping[str, Any]:
    return {k: v.model_dump() for k, v in data.items()}


def serialize_data_mapping(data: DataMapping) -> str:
    return json.dumps(dump_data_mapping(data))


Input_contra = TypeVar("Input_contra", bound=Data, contravariant=True)
Output_co = TypeVar("Output_co", bound=Data, covariant=True)


def get_data_fields(cls: Type[Data]) -> Mapping[str, tuple[ValueType, bool]]:
    """
    Extract the fields of a Data subclass.

    Args:
        cls: The Data subclass to extract fields from

    Returns:
        A mapping of field names to (ValueType, is_required) tuples
    """
    fields: Mapping[str, tuple[ValueType, bool]] = {}
    for k, v in cls.model_fields.items():
        assert v.annotation is not None
        assert issubclass(v.annotation, Value)
        fields[k] = (v.annotation, v.is_required())
    return fields


D = TypeVar("D", bound=Data)


def build_data_type(
    name: str,
    fields: Mapping[str, tuple[ValueType, bool]],
    base_cls: Type[D] = Data,
) -> Type[D]:
    """
    Create a Data subclass whose fields are given by a mapping of field names to
    (ValueType, is_required) tuples.

    This is the inverse of get_fields() - it constructs a class that would return
    the same mapping when passed to get_fields().

    Args:
        name: The name of the class to create
        fields: Mapping of field names to (ValueType, required) tuples
        base_class: The base class to inherit from (defaults to Data)

    Returns:
        A new Pydantic BaseModel class with the specified fields
    """
    # Create field annotations dictionary
    annotations: dict[str, ValueType | tuple[ValueType, Any]] = {
        field_name: value_type if required else (value_type, None)
        for field_name, (value_type, required) in fields.items()
    }

    # Create the class dynamically
    cls = create_model(name, __base__=base_cls, **annotations)  # type: ignore

    return cls


class DataValue(Value[D], Generic[D]):
    """
    A Value subclass that wraps an arbitrary Data object.
    """

    pass


V = TypeVar("V", bound=Value)


@DataValue.register_generic_cast_to(DataValue)
def cast_data_to_data(
    source_type: Type[DataValue],
    target_type: Type[DataValue],
) -> Caster[DataValue, DataValue] | None:
    source_origin, (source_value_type,) = get_origin_and_args(source_type)
    assert source_origin is DataValue
    assert issubclass(source_value_type, Data)

    target_origin, (target_value_type,) = get_origin_and_args(target_type)
    assert target_origin is DataValue
    assert issubclass(target_value_type, Data)

    source_fields = get_data_fields(source_value_type)
    target_fields = get_data_fields(target_value_type)

    for name, (target_field_type, is_required) in target_fields.items():
        if name not in source_fields:
            if is_required:
                return None
            continue

        source_field_type, _ = source_fields[name]
        if not source_field_type.can_cast_to(target_field_type):
            return None

    async def _cast(
        value: source_type,  # pyright: ignore[reportInvalidTypeForm]
        context: "Context",
    ) -> target_type:  # pyright: ignore[reportInvalidTypeForm]
        assert isinstance(value.root, source_value_type)

        items = list(value.root.to_dict().items())
        keys = [k for k, v in items]
        cast_tasks = [v.cast_to(target_fields[k][0], context=context) for k, v in items]
        casted_values = await asyncio.gather(*cast_tasks)
        data_dict = dict(zip(keys, casted_values))
        return target_type(data_dict)

    return _cast


@DataValue.register_generic_cast_to(StringMapValue)
def cast_data_to_string_map(
    source_type: Type[DataValue],
    target_type: Type[StringMapValue[V]],
) -> Caster[DataValue, StringMapValue[V]] | None:
    """
    Casts a DataValue[D] object to a StringMapValue[V] object, if all of the
    fields of D can be cast to V.
    """

    source_origin, (source_value_type,) = get_origin_and_args(source_type)
    assert source_origin is DataValue
    assert issubclass(source_value_type, Data)

    target_origin, (target_value_type,) = get_origin_and_args(target_type)
    assert target_origin is StringMapValue
    assert issubclass(target_value_type, Value)

    source_fields = get_data_fields(source_value_type)
    for source_field_type, _ in source_fields.values():
        if not source_field_type.can_cast_to(target_value_type):
            return None

    async def _cast(
        value: source_type,  # pyright: ignore[reportInvalidTypeForm]
        context: "Context",
    ) -> target_type:  # pyright: ignore[reportInvalidTypeForm]
        assert isinstance(value.root, Data)

        # Cast all fields in parallel
        items = list(value.root.to_dict().items())
        keys = [k for k, v in items]
        cast_tasks = [v.cast_to(target_value_type, context=context) for k, v in items]
        casted_values = await asyncio.gather(*cast_tasks)
        return target_type(dict(zip(keys, casted_values)))  # type: ignore

    return _cast


@StringMapValue.register_generic_cast_to(DataValue)
def cast_string_map_to_data(
    source_type: Type[StringMapValue],
    target_type: Type[DataValue],
) -> Caster[StringMapValue, DataValue] | None:
    """
    Casts a StringMapValue[V] object to a DataValue[D] object by trying to cast
    each value in the map at runtime.

    We don't require statically that V can be cast to the fields of D, because
    in practice V will just be a higher up supertype.
    """

    source_origin, (source_value_type,) = get_origin_and_args(source_type)
    assert source_origin is StringMapValue
    assert issubclass(source_value_type, Value)

    target_origin, (target_value_type,) = get_origin_and_args(target_type)
    assert target_origin is DataValue
    assert issubclass(target_value_type, Data)

    target_fields = get_data_fields(target_value_type)

    for target_field_name, (target_field_type, _) in target_fields.items():
        if not source_value_type.can_cast_to(target_field_type):
            logger.warning(
                "%s to %s: cannot statically cast value type %s to %s (of field %s); will need to rely on a runtime cast which may fail.",
                source_type,
                target_type,
                source_value_type,
                target_field_type,
                target_field_name,
            )

    async def _cast(
        value: source_type,  # pyright: ignore[reportInvalidTypeForm]
        context: "Context",
    ) -> target_type:  # pyright: ignore[reportInvalidTypeForm]
        assert isinstance(value, StringMapValue)

        async def cast_field(
            field_name: str,
            field_value: Value,
        ) -> Value:
            if field_name in target_fields:
                target_field_type, _ = target_fields[field_name]
                casted_value = await field_value.cast_to(
                    target_field_type,
                    context=context,
                )
                return casted_value
            return field_value

        items = list(value.root.items())
        keys = [k for k, v in items]
        cast_tasks = [cast_field(k, v) for k, v in items]
        casted_values = await asyncio.gather(*cast_tasks)
        return target_type(dict(zip(keys, casted_values)))  # type: ignore

    return _cast


__all__ = [
    "build_data_type",
    "Data",
    "DataMapping",
    "DataValue",
    "dump_data_mapping",
    "get_data_fields",
    "Input_contra",
    "Output_co",
    "serialize_data_mapping",
]



# workflow_engine/core/node.py
import asyncio
import logging
import re
import warnings
from collections.abc import Mapping
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    ClassVar,
    Generic,
    Literal,
    Self,
    Type,
    TypeVar,
    Unpack,
    get_origin,
)

from overrides import final
from pydantic import ConfigDict, Field, ValidationError, model_validator

from ..utils.immutable import ImmutableBaseModel
from ..utils.semver import (
    LATEST_SEMANTIC_VERSION,
    SEMANTIC_VERSION_OR_LATEST_PATTERN,
    SEMANTIC_VERSION_PATTERN,
    parse_semantic_version,
)
from .error import NodeException, UserException
from .values import (
    Data,
    DataMapping,
    Value,
    ValueSchema,
    ValueType,
    get_data_fields,
)
from .values.data import Input_contra, Output_co

if TYPE_CHECKING:
    from .context import Context
    from .workflow import Workflow

logger = logging.getLogger(__name__)


class Params(Data):
    model_config: ClassVar[ConfigDict] = ConfigDict(
        extra="allow",
        frozen=True,
    )

    # The base class has extra="allow", so that it can be deserialized into any
    # of its subclasses. However, subclasses should set extra="forbid" to block
    # extra fields.
    def __init_subclass__(cls, **kwargs):
        cls.model_config["extra"] = "forbid"
        super().__init_subclass__(**kwargs)


Params_co = TypeVar("Params_co", bound=Params, covariant=True)
T = TypeVar("T")


@final
class Empty(Params):
    """
    A Data and Params class that is explicitly not allowed to have any
    parameters.
    """

    pass


generic_pattern = re.compile(r"^[a-zA-Z]\w+\[.*\]$")


class NodeTypeInfo(ImmutableBaseModel):
    """
    Information about a node type, in serializable form.
    """

    name: str = Field(
        description="A unique name for the node type, which should be a literal string for concrete subclasses."
    )
    display_name: str = Field(
        description="A human-readable display name for the node, which may or may not be unique."
    )
    description: str | None = Field(
        description="A human-readable description of the node type."
    )
    version: str = Field(
        description="A 3-part version number for the node, following semantic versioning rules (see https://semver.org/).",
        pattern=SEMANTIC_VERSION_PATTERN,
    )
    parameter_schema: ValueSchema = Field(
        default_factory=lambda: Empty.to_value_schema(),
        description="The schema for the parameters of the node type.",
    )
    max_retries: int | None = Field(
        default=None,
        description="Maximum number of retry attempts for this node type. "
        "None means use the execution algorithm's default.",
    )

    @cached_property
    def version_tuple(self) -> tuple[int, int, int]:
        return parse_semantic_version(self.version)

    @classmethod
    def from_parameter_type(
        cls,
        *,
        name: str,
        display_name: str,
        description: str | None = None,
        version: str,
        parameter_type: Type[Params],
        max_retries: int | None = None,
    ) -> Self:
        return cls(
            name=name,
            display_name=display_name,
            description=description,
            version=version,
            parameter_schema=parameter_type.to_value_schema(),
            max_retries=max_retries,
        )


class Node(ImmutableBaseModel, Generic[Input_contra, Output_co, Params_co]):
    """
    A data processing node in a workflow.
    Nodes have three sets of fields:
    - parameter fields must be provided when defining the workflow
    - input fields are provided when executing the workflow
    - output fields are produced by the node if it executes successfully
    """

    # Allow extra fields, such as position or appearance information.
    model_config: ClassVar[ConfigDict] = ConfigDict(extra="allow")

    # Must be annotated as ClassVar[NodeTypeInfo] when overriding.
    # Does not have a value here, since the base Node class is not meant to be
    # instantiated except very temporarily in dispatches.
    TYPE_INFO: ClassVar[NodeTypeInfo]

    type: str = Field(
        description=(
            "The type of the node, which should be a literal string for discriminating concrete subclasses. "
            "Used to determine which node class to load."
        ),
    )
    version: str = Field(
        pattern=SEMANTIC_VERSION_OR_LATEST_PATTERN,
        description=(
            "A 3-part version number for the node, following semantic versioning rules (see https://semver.org/). "
            "There is no guarantee that outdated versions will load successfully. "
            "If not provided, it will default to the current version of the node type."
        ),
        default=LATEST_SEMANTIC_VERSION,
    )
    id: str = Field(
        description="The ID of the node, which must be unique within the workflow."
    )
    params: Params_co = Field(
        default_factory=Empty,  # type: ignore
        description=(
            "Any parameters for the node which are independent of the workflow inputs. "
            "May affect what inputs are accepted by the node."
        ),
    )

    # --------------------------------------------------------------------------
    # SUBCLASS DISPATCH
    # We use this trick to allow Node.model_validate to deserialize nodes as
    # their registered subclasses, using the type field as a discriminator to
    # select the appropriate subclass.
    # Node subclasses are registered automatically when they are defined.

    def __init_subclass__(cls, **kwargs: Unpack[ConfigDict]):
        super().__init_subclass__(**kwargs)  # type: ignore

        # NOTE: something about this hack does not work when using
        # `from __future__ import annotations`.
        while generic_pattern.match(cls.__name__) is not None:
            assert cls.__base__ is not None
            cls = cls.__base__
        type_annotation = cls.__annotations__.get("type", None)
        if type_annotation is None or get_origin(type_annotation) is not Literal:
            _registry.register_base(cls)
        else:
            (type_name,) = type_annotation.__args__
            assert isinstance(type_name, str), type_name
            _registry.register(type_name, cls)

    @model_validator(mode="after")  # type: ignore
    def _to_subclass(self):
        """
        Replaces the Node object with an instance of the registered subclass.
        """
        # HACK: This trick only works if the base class can be instantiated, so
        # we cannot make it an ABC even if it has unimplemented methods.
        if _registry.is_base_class(self.__class__):
            cls = _registry.get(self.type)
            if cls is None:
                raise ValueError(f'Node type "{self.type}" is not registered')
            data = self.model_dump()
            # Attempt migration before dispatching to subclass
            data = _migrate_node_data(data, cls)
            return cls.model_validate(data)
        if self.__class__ is Node:
            warnings.warn(
                f"Node validation for node {self} could not find a registered subclass to dispatch to."
            )
        return self

    # --------------------------------------------------------------------------
    # NAMING

    async def display_name(self, context: "Context") -> str:
        """
        A human-readable display name for the node, which is not necessarily
        unique.
        By default, it is the node type's display name, which is a poor default
        at best.
        You should override this method to provide a more meaningful name and
        disambiguate nodes with the same type.

        This method is async in case determining the node name requires some
        asynchronous work, and can use the context.
        """
        return self.TYPE_INFO.display_name

    def with_namespace(self, namespace: str) -> Self:
        """
        Create a copy of this node with a namespaced ID.

        Args:
            namespace: The namespace to prefix the node ID with

        Returns:
            A new Node with ID '{namespace}/{self.id}'
        """
        return self.model_update(id=f"{namespace}/{self.id}")

    # --------------------------------------------------------------------------
    # VERSIONING

    @property
    def version_tuple(self) -> tuple[int, int, int]:
        """
        The major, minor, and patch version numbers of the node version.
        If the node version is not provided, this will crash.
        """
        return parse_semantic_version(self.version)

    @model_validator(mode="after")
    def validate_version(self):
        """
        Sets the node version to the current version of the node type.
        Validates the node version against the TYPE_INFO version.
        """
        # skip validation for the base Node class, which lacks a TYPE_INFO
        if self.__class__ is Node:
            return self

        type_info = self.__class__.TYPE_INFO
        if self.version == LATEST_SEMANTIC_VERSION:
            self._model_mutate(version=type_info.version)
        elif type_info.version_tuple < self.version_tuple:
            raise ValueError(
                f"Node version {self.version} is newer than the latest version ({type_info.version}) supported by this workflow engine instance."
            )
        elif type_info.version_tuple > self.version_tuple:
            # Migration was attempted in _to_subclass but no migration path exists.
            # Issue a warning but allow the node to load (graceful degradation).
            warnings.warn(
                f"Node version {self.version} is older than the latest version ({type_info.version}) supported by this workflow engine instance, and may need to be migrated."
            )
        return self

    # --------------------------------------------------------------------------
    # TYPING

    @property
    def input_type(self) -> Type[Input_contra]:  # type: ignore (contravariant return type)
        # return Empty to spare users from having to specify the input type on
        # nodes that don't have any input fields
        return Empty  # type: ignore

    @property
    def output_type(self) -> Type[Output_co]:
        # return Empty to spare users from having to specify the output type on
        # nodes that don't have any output fields
        return Empty  # type: ignore

    @property
    def input_fields(self) -> Mapping[str, tuple[ValueType, bool]]:  # type: ignore
        return get_data_fields(self.input_type)

    @property
    def output_fields(self) -> Mapping[str, tuple[ValueType, bool]]:
        return get_data_fields(self.output_type)

    @property
    def input_schema(self) -> ValueSchema:
        return self.input_type.to_value_schema()

    @property
    def output_schema(self) -> ValueSchema:
        return self.output_type.to_value_schema()

    # --------------------------------------------------------------------------
    # EXECUTION

    async def _cast_input(
        self,
        input: DataMapping,
        context: "Context",
    ) -> Input_contra:  # type: ignore (contravariant return type)
        allow_extra_input = (
            self.input_type.model_config.get("extra", "forbid") == "allow"
        )

        # Validate all inputs first
        for key, value in input.items():
            if key not in self.input_fields and allow_extra_input:
                continue
            input_type, _ = self.input_fields[key]
            if not value.can_cast_to(input_type):
                raise UserException(
                    f"Input {value} for node {self.id} is invalid: {value} is not assignable to {input_type}"
                )

        # Cast all inputs in parallel
        cast_tasks: list[Awaitable[Value]] = []
        keys: list[str] = []
        for key, value in input.items():
            if key not in self.input_fields and allow_extra_input:
                continue
            input_type, _ = self.input_fields[key]  # type: ignore
            cast_tasks.append(value.cast_to(input_type, context=context))
            keys.append(key)

        casted_values = await asyncio.gather(*cast_tasks)

        # Build the result dictionary
        casted_input: dict[str, Value] = {}
        for key, casted_value in zip(keys, casted_values):
            casted_input[key] = casted_value

        try:
            return self.input_type.model_validate(casted_input)
        except ValidationError as e:
            raise UserException(
                f"Input {casted_input} for node {self.id} is invalid: {e}"
            )

    # @abstractmethod
    async def run(
        self,
        context: "Context",
        input: Input_contra,
    ) -> "Output_co | Workflow":
        """
        Computes the node's outputs based on its inputs.
        Subclasses must implement this method, but it is not marked as abstract
        because the base Node class needs to be instantiable for dispatching.
        """
        raise NotImplementedError("Subclasses must implement this method")

    @final
    async def __call__(
        self,
        context: "Context",
        input: DataMapping,
    ) -> "DataMapping | Workflow":
        """
        Executes the node.
        """
        try:
            logger.info("Starting node %s", self.id)
            output = await context.on_node_start(node=self, input=input)
            if output is not None:
                return output
            try:
                input_obj = await self._cast_input(input, context)
            except ValidationError as e:
                raise UserException(f"Input {input} for node {self.id} is invalid: {e}")
            output_obj = await self.run(context, input_obj)

            from .workflow import Workflow  # lazy to avoid circular import

            if isinstance(output_obj, Workflow):
                output = output_obj
                # TODO: once that workflow eventually finishes running, its
                # output should be the output of this node, and we should call
                # context.on_node_finish.
            else:
                output = await context.on_node_finish(
                    node=self,
                    input=input,
                    output=output_obj.to_dict(),
                )
            logger.info("Finished node %s", self.id)
            return output
        except Exception as e:
            # In subclasses, you don't have to worry about logging the error,
            # since it'll be logged here.
            logger.exception("Error in node %s", self.id)
            e = await context.on_node_error(node=self, input=input, exception=e)
            if isinstance(e, Mapping):
                logger.exception(
                    "Error absorbed by context and replaced with output %s", e
                )
                return e
            else:
                assert isinstance(e, Exception)
                raise NodeException(self.id) from e


def _migrate_node_data(data: dict[str, Any], target_cls: Type["Node"]) -> dict[str, Any]:
    """
    Attempt to migrate node data to the target class's version.

    This function is called during node deserialization, before the data
    is validated by the target class. If migration is needed and a migration
    path exists, the data is transformed. Otherwise, the original data is
    returned (graceful degradation - validation warnings will be issued later).

    Args:
        data: Raw node data dict with 'type', 'version', 'id', 'params', etc.
        target_cls: The concrete Node subclass to migrate to

    Returns:
        Migrated data dict, or original data if no migration needed/available
    """
    # Skip if target class doesn't have TYPE_INFO (shouldn't happen for concrete classes)
    if not hasattr(target_cls, "TYPE_INFO"):
        return data

    current_version = data.get("version", LATEST_SEMANTIC_VERSION)

    # Skip if using "latest" version (will be resolved to current version later)
    if current_version == LATEST_SEMANTIC_VERSION:
        return data

    type_info = target_cls.TYPE_INFO
    target_version = type_info.version

    # Skip if versions match
    if current_version == target_version:
        return data

    try:
        current_tuple = parse_semantic_version(current_version)
    except ValueError:
        # Invalid version format, let validation handle it
        return data

    # Only migrate if current version is older
    if current_tuple >= type_info.version_tuple:
        return data

    # Attempt migration
    # Import here to avoid circular imports
    from .migration import MigrationNotFoundError, migration_runner

    try:
        migrated_data = migration_runner.migrate(data, target_version)
        logger.debug(
            "Migrated node %s from version %s to %s",
            data.get("id"),
            current_version,
            target_version,
        )
        return migrated_data
    except MigrationNotFoundError:
        # No migration path available - return original data
        # Warning will be issued in validate_version
        logger.debug(
            "No migration path found for node %s from version %s to %s",
            data.get("id"),
            current_version,
            target_version,
        )
        return data


class NodeRegistry:
    def __init__(self):
        self.types: dict[str, Type[Node]] = {}
        self.base_classes: list[Type[Node]] = []

    def register(self, type: str, cls: Type[Node]):
        if type in self.types:
            conflict = self.types[type]
            if cls is not conflict:
                raise ValueError(
                    f'Node type "{type}" (class {cls.__name__}) is already registered to a different class ({conflict.__name__})'
                )
        self.types[type] = cls
        logger.debug("Registering class %s as node type %s", cls.__name__, type)

    def get(self, type: str) -> Type[Node]:
        if type not in self.types:
            raise ValueError(f'Node type "{type}" is not registered')
        return self.types[type]

    def register_base(self, cls: Type[Node]):
        if cls not in self.base_classes:
            self.base_classes.append(cls)
            logger.debug("Registering class %s as base node type", cls.__name__)

    def is_base_class(self, cls: Type[Node]) -> bool:
        return cls in self.base_classes


_registry = NodeRegistry()


__all__ = [
    "Empty",
    "Node",
    "NodeTypeInfo",
    "Params",
]



import pytest

from workflow_engine import Edge, InputEdge, IntegerValue, OutputEdge, Workflow
from workflow_engine.contexts import InMemoryContext
from workflow_engine.execution import TopologicalExecutionAlgorithm
from workflow_engine.nodes import AddNode, ConstantIntegerNode


@pytest.fixture
def workflow():
    """Helper function to create the addition workflow."""
    return Workflow(
        nodes=[
            a := ConstantIntegerNode.from_value(id="a", value=42),
            b := ConstantIntegerNode.from_value(id="b", value=2025),
            a_plus_b := AddNode(id="a+b"),
            a_plus_b_plus_c := AddNode(id="a+b+c"),
        ],
        edges=[
            Edge.from_nodes(
                source=a,
                source_key="value",
                target=a_plus_b,
                target_key="a",
            ),
            Edge.from_nodes(
                source=b,
                source_key="value",
                target=a_plus_b,
                target_key="b",
            ),
            Edge.from_nodes(
                source=a_plus_b,
                source_key="sum",
                target=a_plus_b_plus_c,
                target_key="a",
            ),
        ],
        input_edges=[
            InputEdge.from_node(
                input_key="c",
                target=a_plus_b_plus_c,
                target_key="b",
            ),
        ],
        output_edges=[
            OutputEdge.from_node(
                source=a_plus_b_plus_c,
                source_key="sum",
                output_key="sum",
            ),
        ],
    )


@pytest.mark.unit
def test_workflow_serialization(workflow: Workflow):
    """Test that the workflow can be serialized and deserialized correctly."""
    workflow_json = workflow.model_dump_json(indent=2)
    with open("examples/addition.json", "r") as f:
        assert workflow_json == f.read().strip()

    workflow_json = workflow.model_dump_json()
    deserialized_workflow = Workflow.model_validate_json(workflow_json)
    assert deserialized_workflow == workflow


@pytest.mark.asyncio
async def test_workflow_execution(workflow: Workflow):
    """Test that the workflow executes correctly and produces the expected result."""
    context = InMemoryContext()
    algorithm = TopologicalExecutionAlgorithm()

    c = -256

    errors, output = await algorithm.execute(
        context=context,
        workflow=workflow,
        input={"c": IntegerValue(c)},
    )
    assert not errors.any()
    assert output == {"sum": 42 + 2025 + c}



from hashlib import md5

import pytest

from workflow_engine import (
    Context,
    File,
    IntegerValue,
    SequenceValue,
    StringMapValue,
)
from workflow_engine.contexts.in_memory import InMemoryContext
from workflow_engine.files import JSONFileValue, JSONLinesFileValue


@pytest.fixture
def context():
    """Create a test context for value casting operations."""
    return InMemoryContext()


@pytest.mark.unit
async def test_cast_jsonlines_to_sequence(context: Context):
    """Test that JSONLinesFileValue can be cast to a SequenceValue."""
    jsonl_file = JSONLinesFileValue.from_path("input.jsonl")
    contents = [{"a": 1}, {"b": 2}, {"c": 3}]
    contents_str = '{"a": 1}\n{"b": 2}\n{"c": 3}'

    await jsonl_file.write_data(context, contents)

    assert (await jsonl_file.read_text(context)) == contents_str

    data = await SequenceValue[StringMapValue[IntegerValue]].cast_from(
        jsonl_file,
        context=context,
    )
    assert data == contents

    json_files = await SequenceValue[JSONFileValue].cast_from(
        jsonl_file,
        context=context,
    )
    assert json_files == SequenceValue[JSONFileValue](
        # md5 hashes of the data
        [
            JSONFileValue(File(path=f"{md5(b"{'a': 1}").hexdigest()}.json")),
            JSONFileValue(File(path=f"{md5(b"{'b': 2}").hexdigest()}.json")),
            JSONFileValue(File(path=f"{md5(b"{'c': 3}").hexdigest()}.json")),
        ]
    )
    assert (await json_files[0].read_data(context)) == {"a": 1}
    assert (await json_files[1].read_data(context)) == {"b": 2}
    assert (await json_files[2].read_data(context)) == {"c": 3}



import pytest

from workflow_engine import InputEdge, OutputEdge, Workflow
from workflow_engine.contexts import InMemoryContext
from workflow_engine.execution import TopologicalExecutionAlgorithm
from workflow_engine.nodes import AddNode, ForEachNode


@pytest.fixture
def add_workflow() -> Workflow:
    """Create a workflow that adds two numbers: a + b = c."""
    return Workflow(
        nodes=[
            add := AddNode(id="add"),
        ],
        edges=[],
        input_edges=[
            InputEdge.from_node(
                input_key="a",
                target=add,
                target_key="a",
            ),
            InputEdge.from_node(
                input_key="b",
                target=add,
                target_key="b",
            ),
        ],
        output_edges=[
            OutputEdge.from_node(
                source=add,
                source_key="sum",
                output_key="c",
            ),
        ],
    )


@pytest.fixture
def workflow(add_workflow: Workflow) -> Workflow:
    workflow = Workflow(
        nodes=[
            for_each := ForEachNode.from_workflow(
                id="for_each",
                workflow=add_workflow,
            ),
        ],
        edges=[],
        input_edges=[
            InputEdge.from_node(
                input_key="sequence",
                target=for_each,
                target_key="sequence",
            ),
        ],
        output_edges=[
            OutputEdge.from_node(
                source=for_each,
                source_key="sequence",
                output_key="results",
            ),
        ],
    )

    return workflow


@pytest.mark.asyncio
async def test_for_each_simple_sequence(workflow: Workflow):
    """Test that ForEachNode processes a simple sequence of addition operations."""
    context = InMemoryContext()
    algorithm = TopologicalExecutionAlgorithm()

    input = workflow.input_type.model_validate(
        {
            "sequence": [
                {"a": 1.0, "b": 2.0},
                {"a": 3.0, "b": 4.0},
                {"a": 5.0, "b": 6.0},
            ]
        }
    ).to_dict()

    errors, output = await algorithm.execute(
        context=context,
        workflow=workflow,
        input=input,
    )

    assert not errors.any(), errors

    assert (
        output
        == workflow.output_type.model_validate(
            {
                "results": [
                    {"c": 3.0},
                    {"c": 7.0},
                    {"c": 11.0},
                ]
            }
        ).to_dict()
    )


@pytest.mark.asyncio
async def test_for_each_empty(workflow: Workflow):
    """Test that ForEachNode processes an empty sequence."""
    context = InMemoryContext()
    algorithm = TopologicalExecutionAlgorithm()

    input = workflow.input_type.model_validate(
        {
            "sequence": [],
        }
    ).to_dict()

    errors, output = await algorithm.execute(
        context=context,
        workflow=workflow,
        input=input,
    )

    assert not errors.any(), errors

    assert (
        output
        == workflow.output_type.model_validate(
            {
                "results": [],
            }
        ).to_dict()
    )



from .json import JSONFileValue, JSONLinesFileValue
from .pdf import PDFFileValue
from .text import TextFileValue

__all__ = [
    "JSONFileValue",
    "JSONLinesFileValue",
    "PDFFileValue",
    "TextFileValue",
]



# workflow_engine/nodes/arithmetic.py
"""
Simple nodes for testing the workflow engine, with limited usefulness otherwise.
"""

from typing import ClassVar, Literal

from ..core import (
    Context,
    Data,
    Empty,
    FloatValue,
    IntegerValue,
    Node,
    NodeTypeInfo,
    SequenceValue,
)


class AddNodeInput(Data):
    a: FloatValue
    b: FloatValue


class SumOutput(Data):
    sum: FloatValue


class AddNode(Node[AddNodeInput, SumOutput, Empty]):
    TYPE_INFO: ClassVar[NodeTypeInfo] = NodeTypeInfo(
        name="Add",
        display_name="Add",
        description="Adds two numbers.",
        version="0.4.0",
    )

    type: Literal["Add"] = "Add"  # pyright: ignore[reportIncompatibleVariableOverride]

    @property
    def input_type(self):
        return AddNodeInput

    @property
    def output_type(self):
        return SumOutput

    async def run(self, context: Context, input: AddNodeInput) -> SumOutput:
        return SumOutput(sum=FloatValue(input.a.root + input.b.root))


class SumNodeInput(Data):
    values: SequenceValue[FloatValue]


class SumNodeOutput(Data):
    sum: FloatValue


class SumNode(Node[SumNodeInput, SumNodeOutput, Empty]):
    TYPE_INFO: ClassVar[NodeTypeInfo] = NodeTypeInfo.from_parameter_type(
        name="Sum",
        display_name="Sum",
        description="Sums a sequence of numbers.",
        version="0.4.0",
        parameter_type=Empty,
    )

    type: Literal["Sum"] = "Sum"  # pyright: ignore[reportIncompatibleVariableOverride]

    @property
    def input_type(self):
        return SumNodeInput

    @property
    def output_type(self):
        return SumNodeOutput

    async def run(self, context: Context, input: SumNodeInput) -> SumNodeOutput:
        return SumNodeOutput(sum=FloatValue(sum(v.root for v in input.values)))


class IntegerData(Data):
    value: IntegerValue


class FactorizationData(Data):
    factors: SequenceValue[IntegerValue]


class FactorizationNode(Node[IntegerData, FactorizationData, Empty]):
    TYPE_INFO: ClassVar[NodeTypeInfo] = NodeTypeInfo.from_parameter_type(
        name="Factorization",
        display_name="Factorization",
        description="Factorizes an integer into a sequence of its factors.",
        version="0.4.0",
        parameter_type=Empty,
    )

    type: Literal["Factorization"] = "Factorization"  # pyright: ignore[reportIncompatibleVariableOverride]

    @property
    def input_type(self):
        return IntegerData

    @property
    def output_type(self):
        return FactorizationData

    async def run(self, context: Context, input: IntegerData) -> FactorizationData:
        value = input.value.root
        if value > 0:
            return FactorizationData(
                factors=SequenceValue(
                    [IntegerValue(i) for i in range(1, value + 1) if value % i == 0]
                )
            )
        raise ValueError("Can only factorize positive integers")


__all__ = [
    "AddNode",
    "FactorizationNode",
    "SumNode",
]



# workflow_engine/core/context.py
from abc import ABC, abstractmethod
from typing import TypeVar

from overrides import EnforceOverrides

from .error import ShouldRetry, WorkflowErrors
from .node import Node
from .values import DataMapping, FileValue
from .workflow import Workflow

F = TypeVar("F", bound=FileValue)


class Context(ABC, EnforceOverrides):
    """
    Represents the environment in which a workflow is executed.
    A context's life is limited to the execution of a single workflow.
    """

    @abstractmethod
    async def read(
        self,
        file: FileValue,
    ) -> bytes:
        """
        Read the content of a file from the context.

        file: the file to read

        The context can modify the file by returning a different FileValue.
        """
        raise NotImplementedError("Subclasses must implement this method")

    @abstractmethod
    async def write(
        self,
        file: F,
        content: bytes,
    ) -> F:
        """
        Write the content of a file to the context.

        file: the file to write
        content: the content to write

        The context can modify the file by returning a different FileValue.
        """
        raise NotImplementedError("Subclasses must implement this method")

    async def on_node_start(
        self,
        *,
        node: "Node",
        input: DataMapping,
    ) -> DataMapping | None:
        """
        A hook that is called when a node starts execution.

        If the context already knows what the node's output will be, return that
        output to skip node execution.
        """
        return None

    async def on_node_error(
        self,
        *,
        node: "Node",
        input: DataMapping,
        exception: Exception,
    ) -> Exception | DataMapping:
        """
        A hook that is called when a node raises an error.
        The context can modify the error by returning a different Exception, or
        it can silence the error by returning an output.
        """
        return exception

    async def on_node_retry(
        self,
        *,
        node: "Node",
        input: DataMapping,
        exception: ShouldRetry,
        attempt: int,
    ) -> None:
        """
        A hook that is called when a node is scheduled for retry after raising
        a ShouldRetry exception.

        node: the node that will be retried
        input: the input data to the node
        exception: the ShouldRetry exception that was raised
        attempt: the retry attempt number (1 for first retry, 2 for second, etc.)
        """
        pass

    async def on_node_finish(
        self,
        *,
        node: "Node",
        input: DataMapping,
        output: DataMapping,
    ) -> DataMapping:
        """
        A hook that is called when a node finishes execution by returning a
        DataMapping (not a Workflow).

        node: the node that finished execution
        input: the input data to the node
        output: the output data from the node

        The context can modify the output by returning a different DataMapping.
        """
        return output

    async def on_workflow_start(
        self,
        *,
        workflow: "Workflow",
        input: DataMapping,
    ) -> tuple[WorkflowErrors, DataMapping] | None:
        """
        A hook that is called when a workflow starts execution.

        workflow: the workflow that is starting execution
        input: the input data to the workflow

        If the context already knows what the workflow's output will be, return
        that output to skip workflow execution.
        """
        return None

    async def on_workflow_error(
        self,
        *,
        workflow: "Workflow",
        input: DataMapping,
        errors: WorkflowErrors,
        partial_output: DataMapping,
    ) -> tuple[WorkflowErrors, DataMapping]:
        """
        A hook that is called when a workflow raises an error.

        workflow: the workflow that raised the error
        input: the input data to the workflow
        errors: the errors that occurred
        partial_output: the partial output data from the workflow

        The context can modify the errors or partial output by returning a
        different tuple.
        """
        return errors, partial_output

    async def on_workflow_finish(
        self,
        *,
        workflow: "Workflow",
        input: DataMapping,
        output: DataMapping,
    ) -> DataMapping:
        """
        A hook that is called when a workflow finishes execution.

        workflow: the workflow that finished execution
        input: the input data to the workflow
        output: the output data from the workflow

        The context can modify the output by returning a different DataMapping.
        """
        return output


__all__ = [
    "Context",
]



# tests/test_node.py
from collections.abc import Mapping
from typing import Type

import pytest
from pydantic import ValidationError

from workflow_engine import BooleanValue, Data, IntegerValue, StringValue, ValueType
from workflow_engine.core.values import build_data_type, get_data_fields


@pytest.fixture
def ExampleData() -> Type[Data]:
    """Test data class."""

    class ExampleData(Data):
        name: StringValue
        age: IntegerValue
        active: BooleanValue = None  # type: ignore

    return ExampleData


@pytest.fixture
def example_fields() -> Mapping[str, tuple[ValueType, bool]]:
    """Test fields."""

    return {
        "name": (StringValue, True),
        "age": (IntegerValue, True),
        "active": (BooleanValue, False),
    }


@pytest.mark.unit
def test_get_data_fields(
    ExampleData: Type[Data],
    example_fields: Mapping[str, tuple[ValueType, bool]],
):
    """Test that get_data_fields returns the correct fields."""

    assert get_data_fields(ExampleData) == example_fields


@pytest.mark.unit
def test_build_data_type(
    ExampleData: Type[Data],
    example_fields: Mapping[str, tuple[ValueType, bool]],
):
    """Test that build_data_type returns the correct class."""

    cls = build_data_type("ExampleData", example_fields)

    # can't exactly just test equality, instead we need to test that both
    # classes behave identically in instantiation, serialization, etc.
    assert cls.__name__ == ExampleData.__name__

    # Test that the class can be instantiated
    example_data = {
        "name": "John",
        "age": 30,
        "active": True,
    }

    # Test that each class behaves the same way when instantiating the data
    for kls in (cls, ExampleData):
        instance = kls(name="John", age=30, active=True)  # type: ignore
        deserialized = kls.model_validate(example_data)

        assert instance == deserialized
        assert instance.model_dump() == deserialized.model_dump() == example_data

        # missing optional field
        kls(name="John", age=30)  # type: ignore

        # missing required field
        with pytest.raises(ValidationError):
            kls(name="John", active=True)  # type: ignore

        # bad type
        with pytest.raises(ValidationError):
            kls(name="John", age="thirty", active=True)  # type: ignore

        # extra field
        with pytest.raises(ValidationError):
            kls(name="John", age=30, active=True, extra=1)  # type: ignore



# workflow_engine/nodes/data.py

"""
Utility nodes to construct and deconstruct data objects.
"""

from collections.abc import Sequence
from typing import ClassVar, Generic, Literal, Self, Type, TypeVar

from overrides import override
from pydantic import Field

from workflow_engine.core.values import build_data_type

from ..core import (
    Context,
    Data,
    DataValue,
    Empty,
    IntegerValue,
    Node,
    NodeTypeInfo,
    Params,
    SequenceValue,
    StringMapValue,
    StringValue,
    Value,
    ValueType,
)

V = TypeVar("V", bound=Value)


################################################################################
# Sequences


class SequenceParams(Params):
    length: IntegerValue


class SequenceData(Data, Generic[V]):
    sequence: SequenceValue[V]


class GatherSequenceNode(Node[Data, SequenceData, SequenceParams]):
    """
    Creates a new sequence object of a given length.

    Example:
        >>> node = GatherSequenceNode.from_length("node_id", 3)
        >>> node.run(context, input={}).model_dump()
        {"sequence": [0, 1, 2]}
    """

    TYPE_INFO: ClassVar[NodeTypeInfo] = NodeTypeInfo.from_parameter_type(
        name="GatherSequence",
        display_name="GatherSequence",
        description="Creates a new sequence object of a given length.",
        version="0.4.0",
        parameter_type=SequenceParams,
    )

    type: Literal["GatherSequence"] = "GatherSequence"  # pyright: ignore[reportIncompatibleVariableOverride]

    # The type of the elements in the sequence.
    # For now, this field is only available when the node is constructed
    # programmatically.
    # TODO: make this available at runtime
    element_type: ValueType = Field(default=Value, exclude=True)

    def key(self, index: int) -> str:
        return f"element_{index}"

    @property
    def keys(self) -> Sequence[str]:
        N = self.params.length.root
        return [self.key(i) for i in range(N)]

    @property
    @override
    def input_type(self) -> Type[Data]:
        return build_data_type(
            "GatherSequenceInput",
            {key: (self.element_type, True) for key in self.keys},
        )

    @property
    @override
    def output_type(self) -> Type[SequenceData]:
        return SequenceData[self.element_type]

    @override
    async def run(self, context: Context, input: Data) -> SequenceData:
        input_dict = input.to_dict()
        return self.output_type(
            sequence=SequenceValue[self.element_type](
                root=[input_dict[key] for key in self.keys]
            )
        )

    @classmethod
    def from_length(
        cls,
        id: str,
        length: int,
        element_type: ValueType = Value,
    ) -> Self:
        return cls(
            id=id,
            params=SequenceParams(length=IntegerValue(root=length)),
            element_type=element_type,
        )


class ExpandSequenceNode(Node[SequenceData, Data, SequenceParams]):
    """
    Extracts a sequence of elements to a data object.
    """

    TYPE_INFO: ClassVar[NodeTypeInfo] = NodeTypeInfo.from_parameter_type(
        name="ExpandSequence",
        display_name="ExpandSequence",
        description="Extracts a sequence of elements to a data object.",
        version="0.4.0",
        parameter_type=SequenceParams,
    )

    type: Literal["ExpandSequence"] = "ExpandSequence"  # pyright: ignore[reportIncompatibleVariableOverride]

    # The type of the element to extract.
    # For now, this field is only available when the node is constructed
    # programmatically.
    # TODO: make this available at runtime
    element_type: ValueType = Value

    def key(self, index: int) -> str:
        return f"element_{index}"

    @property
    def keys(self) -> Sequence[str]:
        N = self.params.length.root
        return [self.key(i) for i in range(N)]

    @property
    @override
    def input_type(self) -> Type[SequenceData]:
        return SequenceData[self.element_type]

    @property
    @override
    def output_type(self) -> Type[Data]:
        return build_data_type(
            "ExpandSequenceOutput",
            {key: (self.element_type, True) for key in self.keys},
        )

    @override
    async def run(
        self,
        context: Context,
        input: SequenceData,
    ) -> Data:
        N = self.params.length.root
        assert len(input.sequence) == N, (
            f"Expected sequence of length {N}, but got {len(input.sequence)}"
        )
        return self.output_type(**{self.key(i): input.sequence[i] for i in range(N)})

    @classmethod
    def from_length(
        cls,
        id: str,
        length: int,
        element_type: ValueType = Value,
    ) -> Self:
        return cls(
            id=id,
            params=SequenceParams(length=IntegerValue(root=length)),
            element_type=element_type,
        )


################################################################################
# Mappings


class MappingParams(Params):
    keys: SequenceValue[StringValue]


class MappingData(Data, Generic[V]):
    mapping: StringMapValue[V]


class GatherMappingNode(Node[Data, MappingData, MappingParams]):
    """
    Creates a new mapping object from the inputs to this node.

    Example:
        >>> node = GatherMappingNode.from_keys("node_id", ["a", "b", "c"])
        >>> node.run(context, input={"a": 1, "b": 2, "c": 3}).model_dump()
        {"mapping": {"a": 1, "b": 2, "c": 3}}
    """

    TYPE_INFO: ClassVar[NodeTypeInfo] = NodeTypeInfo.from_parameter_type(
        name="GatherMapping",
        display_name="GatherMapping",
        description="Creates a new mapping object from the inputs to this node.",
        version="0.4.0",
        parameter_type=MappingParams,
    )

    type: Literal["GatherMapping"] = "GatherMapping"  # pyright: ignore[reportIncompatibleVariableOverride]

    # The type of the values in the mapping.
    # For now, this field is only available when the node is constructed
    # programmatically.
    # TODO: make this serializable/deserializable
    value_type: ValueType = Value

    @property
    @override
    def input_type(self) -> Type[Data]:
        return build_data_type(
            "GatherMappingInput",
            {key.root: (self.value_type, True) for key in self.params.keys},
        )

    @property
    @override
    def output_type(self) -> Type[MappingData]:
        return MappingData[self.value_type]

    @override
    async def run(self, context: Context, input: Data) -> MappingData:
        return self.output_type(
            mapping=StringMapValue[self.value_type](
                {key.root: getattr(input, key.root) for key in self.params.keys}
            )
        )

    @classmethod
    def from_keys(
        cls,
        id: str,
        keys: Sequence[str],
    ) -> Self:
        return cls(
            id=id,
            params=MappingParams(
                keys=SequenceValue[StringValue]([StringValue(key) for key in keys])
            ),
        )


class ExpandMappingNode(Node[MappingData, Data, MappingParams]):
    """
    Extracts values from a mapping object at specific keys.

    Example:
        >>> node = ExpandMappingNode.from_keys("node_id", ["a", "b", "c"])
        >>> node.run(context, input={"mapping": {"a": 1, "b": 2, "c": 3}}).model_dump()
        {"a": 1, "b": 2, "c": 3}
    """

    TYPE_INFO: ClassVar[NodeTypeInfo] = NodeTypeInfo.from_parameter_type(
        name="ExpandMapping",
        display_name="ExpandMapping",
        description="Extracts values from a mapping object at specific keys.",
        version="0.4.0",
        parameter_type=MappingParams,
    )

    type: Literal["ExpandMapping"] = "ExpandMapping"  # pyright: ignore[reportIncompatibleVariableOverride]

    # The type of the values in the mapping.
    # For now, this field is only available when the node is constructed
    # programmatically.
    # TODO: make this serializable/deserializable
    value_type: ValueType = Value

    @property
    @override
    def input_type(self) -> Type[MappingData]:
        return MappingData[self.value_type]

    @property
    @override
    def output_type(self) -> Type[Data]:
        return build_data_type(
            "ExpandMappingOutput",
            {key.root: (self.value_type, True) for key in self.params.keys},
        )

    @override
    async def run(self, context: Context, input: MappingData) -> Data:
        return self.output_type(
            **{key.root: input.mapping[key] for key in self.params.keys}
        )

    @classmethod
    def from_keys(cls, id: str, keys: Sequence[str]) -> Self:
        return cls(
            id=id,
            params=MappingParams(
                keys=SequenceValue[StringValue]([StringValue(key) for key in keys])
            ),
        )


################################################################################
# Data nodes


D = TypeVar("D", bound=Data)


class NestedData(Data, Generic[D]):
    """
    A data type that contains a nested data object.
    """

    data: DataValue[D]


class GatherDataNode(Node[Data, NestedData, Empty]):
    """
    A node that gathers a data object from the inputs to this node.

    Example:
        >>> node = GatherDataNode.from_data_type("node_id", Data)
        >>> node.run(context, input={"a": 1, "b": 2, "c": 3}).model_dump()
        {"data": {"a": 1, "b": 2, "c": 3}}
    """

    TYPE_INFO: ClassVar[NodeTypeInfo] = NodeTypeInfo.from_parameter_type(
        name="GatherData",
        display_name="GatherData",
        description="A node that gathers a data object from the inputs to this node.",
        version="0.4.0",
        parameter_type=Empty,
    )

    type: Literal["GatherData"] = "GatherData"  # pyright: ignore[reportIncompatibleVariableOverride]

    # The type of the element to extract.
    # For now, this field is only available when the node is constructed
    # programmatically.
    # TODO: make this serializable/deserializable
    data_type: Type[Data] = Field(default=Data, exclude=True)

    @property
    @override
    def input_type(self) -> Type[Data]:
        return self.data_type

    @property
    @override
    def output_type(self) -> Type[NestedData]:
        return NestedData[self.data_type]

    @override
    async def run(self, context: Context, input: Data) -> NestedData:
        return NestedData[self.data_type](data=DataValue[self.data_type](root=input))

    @classmethod
    def from_data_type(cls, id: str, data_type: Type[Data]) -> Self:
        return cls(
            id=id,
            params=Empty(),
            data_type=data_type,
        )


class ExpandDataNode(Node[NestedData, Data, Empty]):
    """
    A node that expands a nested data object into its individual fields.

    Example:
        >>> node = ExpandDataNode.from_data_type("node_id", Data)
        >>> node.run(context, input={"data": {"a": 1, "b": 2, "c": 3}}).model_dump()
        {"a": 1, "b": 2, "c": 3}
    """

    TYPE_INFO: ClassVar[NodeTypeInfo] = NodeTypeInfo.from_parameter_type(
        name="ExpandData",
        display_name="ExpandData",
        description="A node that expands a nested data object into its individual fields.",
        version="0.4.0",
        parameter_type=Empty,
    )

    type: Literal["ExpandData"] = "ExpandData"  # pyright: ignore[reportIncompatibleVariableOverride]

    # The type of the nested data object.
    # For now, this field is only available when the node is constructed
    # programmatically.
    # TODO: make this serializable/deserializable
    data_type: Type[Data] = Field(default=Data, exclude=True)

    @property
    @override
    def input_type(self) -> Type[NestedData]:
        return NestedData[self.data_type]

    @property
    @override
    def output_type(self) -> Type[Data]:
        return self.data_type

    @override
    async def run(self, context: Context, input: NestedData) -> Data:
        return input.data.root

    @classmethod
    def from_data_type(cls, id: str, data_type: Type[Data]) -> Self:
        return cls(
            id=id,
            params=Empty(),
            data_type=data_type,
        )


__all__ = [
    "ExpandDataNode",
    "ExpandMappingNode",
    "ExpandSequenceNode",
    "GatherDataNode",
    "GatherMappingNode",
    "GatherSequenceNode",
]



# workflow_engine/execution/topological.py
"""
Topological execution algorithm with retry and rate limiting support.
"""

import asyncio

from overrides import override

from ..core import Context, DataMapping, ExecutionAlgorithm, Workflow, WorkflowErrors
from ..core.error import NodeException, ShouldRetry
from .rate_limit import RateLimitRegistry
from .retry import RetryTracker


class TopologicalExecutionAlgorithm(ExecutionAlgorithm):
    """
    Executes the workflow one node at a time on the current thread, in
    topological order.

    Supports retry with backoff for transient failures (ShouldRetry exceptions)
    and rate limiting per node type.
    """

    def __init__(
        self,
        max_retries: int = 3,
        rate_limits: RateLimitRegistry | None = None,
    ):
        """
        Initialize the execution algorithm.

        max_retries: default maximum retry attempts for nodes (can be overridden
                     per node type via NodeTypeInfo.max_retries)
        rate_limits: registry of rate limit configurations per node type
        """
        self.max_retries = max_retries
        self.rate_limits = rate_limits or RateLimitRegistry()

    def _get_node_max_retries(self, node) -> int | None:
        """Get the max retries for a node, checking NodeTypeInfo first."""
        if hasattr(node, "TYPE_INFO") and node.TYPE_INFO.max_retries is not None:
            return node.TYPE_INFO.max_retries
        return None

    @override
    async def execute(
        self,
        *,
        context: Context,
        workflow: Workflow,
        input: DataMapping,
    ) -> tuple[WorkflowErrors, DataMapping]:
        result = await context.on_workflow_start(workflow=workflow, input=input)
        if result is not None:
            return result

        node_outputs: dict[str, DataMapping] = {}
        errors = WorkflowErrors()
        retry_tracker = RetryTracker(default_max_retries=self.max_retries)

        # Track nodes that are waiting for retry (node_id -> input)
        pending_retry: dict[str, DataMapping] = {}

        try:
            ready_nodes = dict(workflow.get_ready_nodes(input=input))

            while ready_nodes or pending_retry:
                # Check if any pending retries are now ready
                for node_id in list(pending_retry.keys()):
                    state = retry_tracker.get_state(node_id)
                    if state.is_ready():
                        ready_nodes[node_id] = pending_retry.pop(node_id)

                # If no nodes are ready, wait for the shortest backoff
                if not ready_nodes and pending_retry:
                    wait_time = retry_tracker.min_wait_time()
                    if wait_time and wait_time.total_seconds() > 0:
                        await asyncio.sleep(wait_time.total_seconds())
                    continue

                if not ready_nodes:
                    break

                node_id, node_input = ready_nodes.popitem()
                node = workflow.nodes_by_id[node_id]

                # Acquire rate limiter if configured for this node type
                limiter = self.rate_limits.get_limiter(node.type)
                if limiter is not None:
                    await limiter.acquire()

                try:
                    node_result = await node(context, node_input)

                    if isinstance(node_result, Workflow):
                        workflow = workflow.expand_node(node_id, node_result)
                    else:
                        node_outputs[node.id] = node_result

                except NodeException as e:
                    # Check if the underlying cause is ShouldRetry
                    if isinstance(e.__cause__, ShouldRetry):
                        should_retry_error = e.__cause__
                        node_max_retries = self._get_node_max_retries(node)

                        if retry_tracker.should_retry(node_id, node_max_retries):
                            retry_tracker.record_retry(node_id, should_retry_error)
                            pending_retry[node_id] = node_input

                            # Call the on_node_retry hook
                            state = retry_tracker.get_state(node_id)
                            await context.on_node_retry(
                                node=node,
                                input=node_input,
                                exception=should_retry_error,
                                attempt=state.attempt,
                            )

                            # Continue to next node without re-raising
                            continue

                    # Max retries exceeded or non-retryable error
                    raise

                finally:
                    if limiter is not None:
                        limiter.release()

                ready_nodes = dict(
                    workflow.get_ready_nodes(
                        input=input,
                        node_outputs=node_outputs,
                        partial_results=ready_nodes,
                    )
                )

            output = workflow.get_output(node_outputs)
        except Exception as e:
            errors.add(e)
            partial_output = workflow.get_output(node_outputs, partial=True)
            errors, partial_output = await context.on_workflow_error(
                workflow=workflow,
                input=input,
                errors=errors,
                partial_output=partial_output,
            )
            return errors, partial_output

        output = await context.on_workflow_finish(
            workflow=workflow,
            input=input,
            output=output,
        )

        return errors, output


__all__ = [
    "TopologicalExecutionAlgorithm",
]



import pytest

from workflow_engine import File, InputEdge, OutputEdge, StringValue, Workflow
from workflow_engine.contexts import InMemoryContext
from workflow_engine.execution import TopologicalExecutionAlgorithm
from workflow_engine.files import TextFileValue
from workflow_engine.nodes import AppendToFileNode


@pytest.fixture
def workflow():
    """Helper function to create the append workflow."""
    return Workflow(
        nodes=[
            append := AppendToFileNode.from_suffix(
                id="append",
                suffix="_append",
            ),
        ],
        edges=[],
        input_edges=[
            InputEdge(input_key="text", target_id=append.id, target_key="text"),
            InputEdge(input_key="file", target_id=append.id, target_key="file"),
        ],
        output_edges=[
            OutputEdge(source_id=append.id, source_key="file", output_key="file"),
        ],
    )


@pytest.mark.unit
def test_workflow_serialization(workflow: Workflow):
    """Test that the append workflow can be serialized and deserialized correctly."""
    workflow_json = workflow.model_dump_json(indent=2)
    with open("examples/append.json", "r") as f:
        assert workflow_json == f.read().strip()

    workflow_json = workflow.model_dump_json()
    deserialized_workflow = Workflow.model_validate_json(workflow_json)
    assert deserialized_workflow == workflow


@pytest.mark.asyncio
async def test_workflow_execution(workflow: Workflow):
    """Test that the workflow executes correctly and produces the expected result."""
    context = InMemoryContext()
    algorithm = TopologicalExecutionAlgorithm()

    # Create input with a text file
    hello_world = "Hello, world!"
    input_file = TextFileValue(File(path="test.txt"))
    input_file = await input_file.write_text(context, text=hello_world)

    appended_text = StringValue("This text will be appended to the file.")
    errors, output = await algorithm.execute(
        context=context,
        workflow=workflow,
        input={
            "file": input_file,
            "text": appended_text,
        },
    )

    # Verify no errors occurred
    assert not errors.any()

    # Verify the output file exists and has the correct content
    output_file = output["file"]
    assert isinstance(output_file, TextFileValue)
    assert output_file.path == "test_append.txt"
    output_text = await output_file.read_text(context)
    assert output_text == hello_world + appended_text.root



# workflow_engine/core/values/__init__.py
from .data import (
    Data,
    DataMapping,
    DataValue,
    build_data_type,
    dump_data_mapping,
    get_data_fields,
    serialize_data_mapping,
)
from .file import File, FileValue
from .json import JSON, JSONValue
from .mapping import StringMapValue
from .primitives import BooleanValue, FloatValue, IntegerValue, NullValue, StringValue
from .schema import ValueSchema, ValueSchemaValue, validate_value_schema
from .sequence import SequenceValue
from .value import Caster, Value, ValueType, get_origin_and_args

__all__ = [
    "BooleanValue",
    "build_data_type",
    "Caster",
    "Data",
    "DataMapping",
    "DataValue",
    "dump_data_mapping",
    "File",
    "FileValue",
    "FloatValue",
    "get_data_fields",
    "get_origin_and_args",
    "IntegerValue",
    "JSON",
    "JSONValue",
    "NullValue",
    "SequenceValue",
    "serialize_data_mapping",
    "StringMapValue",
    "StringValue",
    "validate_value_schema",
    "Value",
    "ValueSchema",
    "ValueSchemaValue",
    "ValueType",
]



# workflow_engine/core/execution.py
from abc import ABC, abstractmethod

from overrides import EnforceOverrides

from .context import Context
from .error import WorkflowErrors
from .values import DataMapping
from .workflow import Workflow


class ExecutionAlgorithm(ABC, EnforceOverrides):
    """
    Handles the scheduling and execution of workflow nodes.
    Uses hooks to perform extra functionality at key points in the execution
    flow.
    """

    @abstractmethod
    async def execute(
        self,
        *,
        context: Context,
        workflow: Workflow,
        input: DataMapping,
    ) -> tuple[WorkflowErrors, DataMapping]:
        pass


__all__ = [
    "ExecutionAlgorithm",
]



# workflow_engine/core/edge.py

from ..utils.immutable import ImmutableBaseModel
from .node import Node
from .values import Value, ValueType


class Edge(ImmutableBaseModel):
    """
    An edge connects the output of source node to the input of a target node.
    """

    source_id: str
    source_key: str
    target_id: str
    target_key: str

    @classmethod
    def from_nodes(
        cls,
        *,
        source: Node,
        source_key: str,
        target: Node,
        target_key: str,
    ) -> "Edge":
        """
        Self-validating factory method.
        """
        edge = cls(
            source_id=source.id,
            source_key=source_key,
            target_id=target.id,
            target_key=target_key,
        )
        edge.validate_types(source, target)
        return edge

    def validate_types(self, source: Node, target: Node):
        if self.source_key not in source.output_fields:
            raise ValueError(
                f"Source node {source.id} does not have a {self.source_key} field"
            )

        if self.target_key not in target.input_fields:
            raise ValueError(
                f"Target node {target.id} does not have a {self.target_key} field"
            )

        source_output_type, _ = source.output_fields[self.source_key]
        assert issubclass(source_output_type, Value)
        target_input_type, _ = target.input_fields[self.target_key]
        assert issubclass(target_input_type, Value)

        if not source_output_type.can_cast_to(target_input_type):
            raise TypeError(
                f"Edge from {source.id}.{self.source_key} to {target.id}.{self.target_key} has invalid types: {source_output_type} is not assignable to {target_input_type}"
            )


class SynchronizationEdge(ImmutableBaseModel):
    """
    An edge that carries no information, but indicates that the target node must
    run after the source node finishes.
    """

    source_id: str
    target_id: str


class InputEdge(ImmutableBaseModel):
    """
    An "edge" that maps a field from the workflow's input to the input of a
    target node.
    """

    input_key: str
    target_id: str
    target_key: str

    @classmethod
    def from_node(
        cls,
        *,
        input_key: str,
        target: Node,
        target_key: str,
    ) -> "InputEdge":
        return cls(
            input_key=input_key,
            target_id=target.id,
            target_key=target_key,
        )

    def validate_types(self, input_type: ValueType, target: Node):
        if self.target_key not in target.input_fields:
            raise ValueError(
                f"Target node {target.id} does not have a {self.target_key} field"
            )

        target_input_type, _ = target.input_fields[self.target_key]
        assert issubclass(target_input_type, Value)

        if not input_type.can_cast_to(target_input_type):
            raise TypeError(
                f"Input edge to {target.id}.{self.target_key} has invalid types: {input_type} is not assignable to {target_input_type}"
            )


class OutputEdge(ImmutableBaseModel):
    """
    An "edge" that maps a source node's output to a special output of the
    workflow.
    """

    source_id: str
    source_key: str
    output_key: str

    @classmethod
    def from_node(
        cls,
        *,
        source: Node,
        source_key: str,
        output_key: str,
    ) -> "OutputEdge":
        return cls(
            source_id=source.id,
            source_key=source_key,
            output_key=output_key,
        )

    def validate_types(self, source: Node, output_type: ValueType):
        if self.source_key not in source.output_fields:
            raise ValueError(
                f"Source node {source.id} does not have a {self.source_key} field"
            )

        source_output_type, _ = source.output_fields[self.source_key]
        assert issubclass(source_output_type, Value)

        if not source_output_type.can_cast_to(output_type):
            raise TypeError(
                f"Output edge from {source.id}.{self.source_key} has invalid types: {source_output_type} is not assignable to {output_type}"
            )


__all__ = [
    "Edge",
    "InputEdge",
    "OutputEdge",
    "SynchronizationEdge",
]



# workflow_engine/core/values/json.py

from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING

from .value import Value

if TYPE_CHECKING:
    from ..context import Context


type JSON = Mapping[str, JSON] | Sequence[JSON] | None | bool | int | float | str


class JSONValue(Value[JSON]):
    pass


@Value.register_cast_to(JSONValue)
def cast_any_to_json(value: Value, context: "Context") -> JSONValue:
    return JSONValue(value.model_dump())


__all__ = [
    "JSON",
    "JSONValue",
]



# workflow_engine/core/values/schema.py
"""
Pydantic models that cover the subset of the JSON Schema language generated by
Pydantic for Value subclasses.
"""

from __future__ import annotations

from collections.abc import Mapping, Sequence
from functools import cached_property
from typing import Any, ClassVar, Literal

from overrides import override
from pydantic import (
    ConfigDict,
    Field,
    ValidationError,
    model_serializer,
    SerializerFunctionWrapHandler,
    model_validator,
)

from ...utils.immutable import ImmutableBaseModel
from .data import Data, DataValue, build_data_type
from .mapping import StringMapValue
from .primitives import (
    BooleanValue,
    FloatValue,
    IntegerValue,
    NullValue,
    StringValue,
)
from .sequence import SequenceValue
from .value import Value, ValueType, value_type_registry


def merge_defs(
    a: Mapping[str, ValueSchema] | None,
    b: Mapping[str, ValueSchema] | None,
) -> Mapping[str, ValueSchema] | None:
    if a is None:
        return {} if b is None else b
    if b is None:
        return {} if a is None else a
    return dict(**a, **b)


class BaseValueSchema(ImmutableBaseModel):
    # We will only handle fields that Pydantic actually uses.
    # extra="forbid" offensively validates that we haven't missed any fields.
    model_config: ClassVar[ConfigDict] = ConfigDict(
        extra="forbid",
        serialize_by_alias=True,
    )

    # NOTE: recursive validation does not work unless this is a dict instead of a Mapping
    defs: dict[str, ValueSchema] = Field(
        default_factory=dict,
        serialization_alias="$defs",
        validation_alias="$defs",
    )
    title: str | None = None
    description: str | None = None
    default: Any | None = None
    const: Any | None = None

    @model_validator(mode="before")
    @classmethod
    def rename_defs_alias(cls, data: Any) -> Any:
        """
        A Pydantic bug causes model_validate() to break when called on an
        instance of BaseValueSchema.
        This is because the validator receives the .defs property from the
        object, whereas it is expecting $defs.
        To fix this, we move the property to the expected key.
        """
        if isinstance(data, Mapping) and ("$defs" not in data) and ("defs" in data):
            data = dict(data)
            data["$defs"] = data.pop("defs")
        return data

    @model_serializer(mode="wrap")
    def serialize_without_unset(self, handler: SerializerFunctionWrapHandler):
        """
        Custom serializer that excludes unset fields by default, so we don't
        clutter the output with unused fields.
        Applies to all subclasses of BaseValueSchema.

        Newbies might do this by overriding .model_dump(), however this is a
        trap because .model_dump() is not called recursively when this model is
        nested inside another model.

        Has special handling for $defs (included as long as it is not empty)
        and $ref (always included).
        """
        # Get the serialized data from the handler
        data: dict[str, Any] = handler(self)

        keys_to_keep: set[str] = {"$ref"}  # always keep $ref if present

        if hasattr(self, "__pydantic_fields_set__"):
            keys_to_keep.update(self.__pydantic_fields_set__)

        # special handling for $defs, because aliasing breaks
        # __pydantic_fields_set__
        if "$defs" in data and not (
            isinstance(data["$defs"], Mapping) and len(data["$defs"]) == 0
        ):
            keys_to_keep.add("$defs")

        return {k: v for k, v in data.items() if k in keys_to_keep}

    def to_value_cls(
        self,
        *extra_defs: Mapping[str, ValueSchema],
    ) -> ValueType:
        """
        Resolves this schema to a Pydantic class.
        If the title matches a concrete Value class, then it returns that class
        right away, regardless the other contents of the schema.
        This is technically a hack, but it is a massive quality of life boost
        for users, who can just provide { "title": "CustomValue" } as an alias
        for the entire CustomValue schema.

        References, if any, are resolved using self.defs first, then any
        extra_defs in order of decreasing precedence.
        """
        if self.title is not None and self.title in value_type_registry:
            return value_type_registry[self.title]
        return self.build_value_cls(*extra_defs)

    def build_value_cls(
        self,
        *extra_defs: Mapping[str, ValueSchema],
    ) -> ValueType:
        """
        Builds a Pydantic class from this schema. References, if any, are
        resolved using self.defs first, then any extra_defs in order of decreasing precedence.
        """
        raise NotImplementedError("Subclasses must implement this method")


class BooleanValueSchema(BaseValueSchema):
    type: Literal["boolean"]

    @override
    def build_value_cls(
        self,
        *extra_defs: Mapping[str, ValueSchema],
    ) -> type[BooleanValue]:
        return BooleanValue


class NullValueSchema(BaseValueSchema):
    type: Literal["null"]

    @override
    def build_value_cls(
        self,
        *extra_defs: Mapping[str, ValueSchema],
    ) -> type[NullValue]:
        return NullValue


class IntegerValueSchema(BaseValueSchema):
    # unused JSON Schema fields include minimum, maximum, multipleOf, etc.
    type: Literal["integer"]

    @override
    def build_value_cls(
        self,
        *extra_defs: Mapping[str, ValueSchema],
    ) -> type[IntegerValue]:
        return IntegerValue


class FloatValueSchema(BaseValueSchema):
    # unused JSON Schema fields include minimum, maximum, multipleOf, etc.
    type: Literal["number"]

    @override
    def build_value_cls(
        self,
        *extra_defs: Mapping[str, ValueSchema],
    ) -> type[FloatValue]:
        return FloatValue


class StringValueSchema(BaseValueSchema):
    # unused JSON Schema fields include length
    type: Literal["string"]
    enum: Sequence[str] | None = None
    pattern: str | None = None

    @override
    def build_value_cls(
        self,
        *extra_defs: Mapping[str, ValueSchema],
    ) -> type[StringValue]:
        return StringValue


class SequenceValueSchema(BaseValueSchema):
    type: Literal["array"]
    items: ValueSchema

    @override
    def build_value_cls(
        self,
        *extra_defs: Mapping[str, ValueSchema],
    ) -> type[SequenceValue]:
        T = self.items.to_value_cls(self.defs, *extra_defs)
        return SequenceValue[T]


class StringMapValueSchema(BaseValueSchema):
    type: Literal["object"]
    additionalProperties: ValueSchema | Literal[True] = True

    @override
    def build_value_cls(
        self,
        *extra_defs: Mapping[str, ValueSchema],
    ) -> type[StringMapValue]:
        if self.additionalProperties is True:
            return StringMapValue[Value]
        else:
            V = self.additionalProperties.to_value_cls(self.defs, *extra_defs)
            return StringMapValue[V]


class DataValueSchema(BaseValueSchema):
    """
    Matches a DataValue[T] schema, for some class T that inherits from Data.
    """

    type: Literal["object"]

    # NOTE: recursive validation does not work unless this is a dict instead of a Mapping
    properties: dict[str, ValueSchema]
    additionalProperties: ValueSchema | bool = False
    required: Sequence[str] = Field(default_factory=list)

    def build_data_cls(
        self,
        *extra_defs: Mapping[str, ValueSchema],
    ) -> type[Data]:
        """
        Rebuilds the Data type T from the schema.
        This rebuilt Data type will not be the same as the original Data type T
        (i.e. equality checks will fail), but it will have the same fields.
        It will not necessarily have the same to_value_schema() as T, because
        the rebuilding process does not yet capture all of the descriptions from
        the original schema.
        """
        # TODO: capture the descriptions from the schema
        required_set = frozenset(self.required)
        properties = {
            k: (
                v.to_value_cls(self.defs, *extra_defs),
                k in required_set,
            )
            for k, v in self.properties.items()
        }
        assert self.title is not None
        return build_data_type(self.title, properties)

    @override
    def build_value_cls(
        self,
        *extra_defs: Mapping[str, ValueSchema],
    ) -> type[DataValue]:
        D = self.build_data_cls(*extra_defs)
        return DataValue[D]


class UnionValueSchema(BaseValueSchema):
    anyOf: Sequence[ValueSchema]

    @override
    def build_value_cls(
        self,
        *extra_defs: Mapping[str, ValueSchema],
    ) -> ValueType:
        # TODO: add proper type union support
        raise NotImplementedError(
            "There is no Value type handling for type unions yet."
        )


class ReferenceValueSchema(BaseValueSchema):
    """
    A schema of the form
    { "$defs": { "foo": ... }, "$ref": "#/$defs/foo" }
    """

    model_config: ClassVar[ConfigDict] = ConfigDict(
        serialize_by_alias=True,  # already set in BaseValueSchema, but setting again here to be explicit
    )

    ref: str = Field(
        pattern=r"^#/\$defs/\w+$",
        serialization_alias="$ref",
        validation_alias="$ref",
    )

    @model_validator(mode="before")
    @classmethod
    def rename_ref_alias(cls, data: Any) -> Any:
        """
        A Pydantic bug causes model_validate() to break when called on an
        instance of ReferenceValueSchema.
        This is because the validator receives the .ref property from the
        object, whereas it is expecting $ref.
        To fix this, we move the property to the expected key.
        """
        if isinstance(data, Mapping) and ("$ref" not in data) and ("ref" in data):
            data = dict(data)
            data["$ref"] = data.pop("ref")
            return data
        return data

    @cached_property
    def id(self) -> str:
        _hash, _defs, id = self.ref.split("/")
        assert _hash == "#"
        assert _defs == "$defs"
        return id

    @override
    def build_value_cls(
        self,
        *extra_defs: Mapping[str, ValueSchema],
    ) -> ValueType:
        for defs in (self.defs, *extra_defs):
            if self.id in defs:
                return defs[self.id].to_value_cls(self.defs, *extra_defs)
        raise KeyError(f"Schema definition for {self.id} not found")


type ValueSchema = (
    BooleanValueSchema
    | DataValueSchema
    | FloatValueSchema
    | IntegerValueSchema
    | NullValueSchema
    | SequenceValueSchema
    | StringMapValueSchema
    | StringValueSchema
    | UnionValueSchema
    | ReferenceValueSchema  # must be handled second-to-last
    | BaseValueSchema  # must be handled last, corresponds to the Any catch-all
)


# yep, this is possible
class ValueSchemaValue(Value[ValueSchema]):
    pass


def validate_value_schema(schema: Any) -> ValueSchema:
    # idempotent: if the schema is already a ValueSchema, just return it
    if isinstance(schema, BaseValueSchema):
        return schema
    try:
        return ValueSchemaValue.model_validate(schema).root
    except ValidationError as e:
        raise ValueError(f"Invalid value schema: {schema}") from e


__all__ = [
    "ValueSchema",
    "ValueSchemaValue",
]



# workflow_engine/core/workflow.py
from collections.abc import Mapping, Sequence
from functools import cached_property
from itertools import chain
from typing import Type

import networkx as nx
from pydantic import ConfigDict, ValidationError, model_validator

from ..utils.immutable import ImmutableBaseModel
from .edge import Edge, InputEdge, OutputEdge
from .error import NodeExpansionException, UserException
from .node import Node
from .values import Data, DataMapping, Value, ValueType, build_data_type


class Workflow(ImmutableBaseModel):
    model_config = ConfigDict(frozen=True)

    nodes: Sequence[Node]
    edges: Sequence[Edge]
    input_edges: Sequence[InputEdge]
    output_edges: Sequence[OutputEdge]

    @cached_property
    def nodes_by_id(self) -> Mapping[str, Node]:
        nodes_by_id: dict[str, Node] = {}
        for node in self.nodes:
            if node.id in nodes_by_id:
                raise ValueError(f"Node {node.id} is already in the graph")
            nodes_by_id[node.id] = node
        return nodes_by_id

    @cached_property
    def edges_by_target(self) -> Mapping[str, Mapping[str, Edge | InputEdge]]:
        """
        A mapping from each node and input key to the (unique) edge that targets
        the node at that key.
        """
        edges_by_target: dict[str, dict[str, Edge | InputEdge]] = {
            node.id: {} for node in self.nodes
        }
        for edge in chain(self.edges, self.input_edges):
            if edge.target_key in edges_by_target[edge.target_id]:
                raise ValueError(
                    f"In-edge to {edge.target_id}.{edge.target_key} is already in the graph"
                )
            edges_by_target[edge.target_id][edge.target_key] = edge
        return edges_by_target

    @cached_property
    def input_fields(self) -> Mapping[str, tuple[ValueType, bool]]:
        return {
            edge.input_key: self.nodes_by_id[edge.target_id].input_fields[
                edge.target_key
            ]
            for edge in self.input_edges
        }

    @cached_property
    def output_fields(self) -> Mapping[str, tuple[ValueType, bool]]:
        return {
            edge.output_key: self.nodes_by_id[edge.source_id].output_fields[
                edge.source_key
            ]
            for edge in self.output_edges
        }

    @cached_property
    def input_type(self) -> Type[Data]:
        return build_data_type(
            "WorkflowInput",
            {
                edge.input_key: self.input_fields[edge.input_key]
                for edge in self.input_edges
            },
        )

    @cached_property
    def output_type(self) -> Type[Data]:
        return build_data_type(
            "WorkflowOutput",
            {
                edge.output_key: self.output_fields[edge.output_key]
                for edge in self.output_edges
            },
        )

    @cached_property
    def nx_graph(self) -> nx.DiGraph:
        graph = nx.DiGraph()

        for node in self.nodes:
            graph.add_node(node.id, data=node)

        for edge in self.edges:
            graph.add_edge(edge.source_id, edge.target_id, data=edge)

        return graph

    @cached_property
    def input_edges_by_key(self) -> Mapping[str, InputEdge]:
        """Index of input edges by their input_key."""
        return {edge.input_key: edge for edge in self.input_edges}

    @cached_property
    def output_edges_by_key(self) -> Mapping[str, OutputEdge]:
        """Index of output edges by their output_key."""
        return {edge.output_key: edge for edge in self.output_edges}

    @model_validator(mode="after")
    def _validate_nodes(self):
        # make sure that for each node, all input edges are present
        for node in self.nodes:
            for key, (_, required) in node.input_fields.items():
                if required and key not in self.edges_by_target[node.id]:
                    raise ValueError(f"Node {node.id} has no required input edge {key}")
        return self

    @model_validator(mode="after")
    def _validate_edges(self):
        for edge in self.edges:
            edge.validate_types(
                source=self.nodes_by_id[edge.source_id],
                target=self.nodes_by_id[edge.target_id],
            )
        return self

    @model_validator(mode="after")
    def _validate_dag(self):
        if not nx.is_directed_acyclic_graph(self.nx_graph):
            cycles = list(nx.simple_cycles(self.nx_graph))
            raise ValueError(f"Workflow graph is not a DAG. Cycles found: {cycles}")
        return self

    @model_validator(mode="after")
    def _validate_no_id_prefix_collisions(self):
        """
        Ensure no node ID is a prefix of another when followed by '/'.

        This prevents ID collisions when composite nodes are expanded.
        For example, this prevents having both 'foo' and 'foo/bar' nodes.
        """
        node_ids = [node.id for node in self.nodes]
        sorted_ids = sorted(node_ids)

        for i in range(len(sorted_ids) - 1):
            current = sorted_ids[i]
            next_id = sorted_ids[i + 1]
            if next_id.startswith(current + "/"):
                raise ValueError(
                    f"Node ID collision detected: '{current}' is a prefix of '{next_id}'. "
                    f"This would cause conflicts when composite nodes are expanded. "
                    f"Please ensure no node ID is a prefix of another when followed by '/'."
                )
        return self

    def get_ready_nodes(
        self,
        input: DataMapping,
        node_outputs: Mapping[str, DataMapping] | None = None,
        partial_results: Mapping[str, DataMapping] | None = None,
    ) -> Mapping[str, DataMapping]:
        """
        Given the input and the set of nodes which have already finished, return
        the nodes that are now ready to be executed and their arguments.

        Should only return an empty map if the entire workflow is finished.

        For efficiency, this method can use partial results to avoid
        recalculating already finished nodes.
        """
        if node_outputs is None:
            node_outputs = {}

        ready_nodes: dict[str, DataMapping] = (
            {} if partial_results is None else dict(partial_results)
        )
        for node in self.nodes:
            # remove the node if it is now finished
            if node.id in node_outputs:
                if node.id in ready_nodes:
                    ready_nodes.pop(node.id)
                continue
            # skip the node if it is already in the ready set
            if node.id in ready_nodes:
                continue

            # node might be ready, we have to check all its input edges
            ready: bool = True
            node_input_dict: DataMapping = {}
            for target_key, edge in self.edges_by_target[node.id].items():
                # if the input is missing, we will let the node figure it out
                if isinstance(edge, InputEdge):
                    node_input_dict[target_key] = input[edge.input_key]
                elif edge.source_id in node_outputs:
                    node_input_dict[target_key] = node_outputs[edge.source_id][
                        edge.source_key
                    ]
                else:
                    ready = False
                    break
            if not ready:
                continue

            try:
                ready_nodes[node.id] = node_input_dict
            except ValidationError as e:
                raise UserException(
                    f"Input {node_input_dict} for node {node.id} is invalid: {e}",
                )
        return ready_nodes

    def get_output(
        self,
        node_outputs: Mapping[str, DataMapping],
        partial: bool = False,
    ) -> DataMapping:
        """
        Get the output of the workflow.

        If partial is True, this method should never raise an exception, and the
        output will only include nodes that have been executed, for which the
        output field is available.
        """
        output: DataMapping = {}
        for edge in self.output_edges:
            if edge.source_id not in node_outputs:
                if partial:
                    continue
                raise UserException(
                    f"Cannot get output from node {edge.source_id}.",
                )
            node_output = node_outputs[edge.source_id]
            if edge.source_key not in node_output:
                if partial:
                    continue
                raise UserException(
                    f"Cannot get output from node {edge.source_id} at key '{edge.source_key}'.",
                )
            output_field = node_output[edge.source_key]
            output[edge.output_key] = output_field
        return output

    def expand_node(
        self,
        node_id: str,
        subgraph: "Workflow",
    ) -> "Workflow":
        """
        Replace a node in this workflow with a subgraph.

        This method performs graph surgery to replace a node with a subgraph.
        The subgraph's nodes are namespaced with the original node's ID to prevent
        ID collisions. Input and output edges are reconnected appropriately.

        Args:
            node_id: ID of the node to replace
            subgraph: The workflow to insert in place of the node
            node_input: The input that was passed to the original node

        Returns:
            A new Workflow with the node replaced by the subgraph

        Raises:
            ValueError: If the node_id doesn't exist or if the replacement would
                       create an invalid graph
        """
        try:
            if node_id not in self.nodes_by_id:
                raise ValueError(f"Node {node_id} not found in workflow")

            subgraph = subgraph.with_namespace(node_id)

            # Collect all edges that need to be modified
            new_nodes: list[Node] = [
                node for node in self.nodes if node.id != node_id
            ] + list(subgraph.nodes)
            new_edges: list[Edge] = list(subgraph.edges)
            new_input_edges: list[InputEdge] = []
            new_output_edges: list[OutputEdge] = []

            # Use cached properties for subgraph edge indexing
            subgraph_input_by_key = subgraph.input_edges_by_key
            subgraph_output_by_key = subgraph.output_edges_by_key

            # Handle input edges - reconnect them to subgraph input nodes
            for input_edge in self.input_edges:
                if input_edge.target_id == node_id:
                    if input_edge.target_key in subgraph_input_by_key:
                        subgraph_input_edge = subgraph_input_by_key[
                            input_edge.target_key
                        ]
                        new_input_edges.append(
                            InputEdge(
                                input_key=input_edge.input_key,
                                target_id=subgraph_input_edge.target_id,
                                target_key=subgraph_input_edge.target_key,
                            )
                        )
                    # If no matching input edge found, the edge is dropped
                else:
                    new_input_edges.append(input_edge)

            # Handle regular edges
            for edge in self.edges:
                if edge.target_id == node_id:
                    if edge.target_key in subgraph_input_by_key:
                        subgraph_input_edge = subgraph_input_by_key[edge.target_key]
                        new_edges.append(
                            Edge(
                                source_id=edge.source_id,
                                source_key=edge.source_key,
                                target_id=subgraph_input_edge.target_id,
                                target_key=subgraph_input_edge.target_key,
                            )
                        )
                    # If no matching input edge found, the edge is dropped
                elif edge.source_id == node_id:
                    if edge.source_key in subgraph_output_by_key:
                        subgraph_output_edge = subgraph_output_by_key[edge.source_key]
                        new_edges.append(
                            Edge(
                                source_id=subgraph_output_edge.source_id,
                                source_key=subgraph_output_edge.source_key,
                                target_id=edge.target_id,
                                target_key=edge.target_key,
                            )
                        )
                    else:
                        raise ValueError(
                            f"Node {node_id} has output '{edge.source_key}' that is required by "
                            f"workflow, but the subgraph does not provide a matching output edge"
                        )
                else:
                    new_edges.append(edge)

            # Handle output edges
            for output_edge in self.output_edges:
                if output_edge.source_id == node_id:
                    if output_edge.source_key in subgraph_output_by_key:
                        subgraph_output_edge = subgraph_output_by_key[
                            output_edge.source_key
                        ]
                        new_output_edges.append(
                            OutputEdge(
                                source_id=subgraph_output_edge.source_id,
                                source_key=subgraph_output_edge.source_key,
                                output_key=output_edge.output_key,
                            )
                        )
                    else:
                        raise ValueError(
                            f"Node {node_id} has output '{output_edge.source_key}' that is required by "
                            f"workflow output '{output_edge.output_key}', but the subgraph does not "
                            f"provide a matching output edge"
                        )
                else:
                    new_output_edges.append(output_edge)

            return Workflow(
                nodes=new_nodes,
                edges=new_edges,
                input_edges=new_input_edges,
                output_edges=new_output_edges,
            )
        except Exception as e:
            raise NodeExpansionException(node_id, workflow=subgraph) from e

    def with_namespace(self, namespace: str) -> "Workflow":
        """
        Create a copy of this workflow with all node IDs namespaced.

        Args:
            namespace: The namespace to prefix all node IDs with

        Returns:
            A new Workflow with all node IDs prefixed with '{namespace}/'
        """
        # Create namespaced nodes
        namespaced_nodes = [node.with_namespace(namespace) for node in self.nodes]

        # Create namespaced edges (update source_id and target_id)
        namespaced_edges = [
            edge.model_copy(
                update={
                    "source_id": f"{namespace}/{edge.source_id}",
                    "target_id": f"{namespace}/{edge.target_id}",
                }
            )
            for edge in self.edges
        ]

        # Create namespaced input edges (update target_id only)
        namespaced_input_edges = [
            input_edge.model_copy(
                update={"target_id": f"{namespace}/{input_edge.target_id}"}
            )
            for input_edge in self.input_edges
        ]

        # Create namespaced output edges (update source_id only)
        namespaced_output_edges = [
            output_edge.model_copy(
                update={"source_id": f"{namespace}/{output_edge.source_id}"}
            )
            for output_edge in self.output_edges
        ]

        return Workflow(
            nodes=namespaced_nodes,
            edges=namespaced_edges,
            input_edges=namespaced_input_edges,
            output_edges=namespaced_output_edges,
        )


class WorkflowValue(Value[Workflow]):
    pass


__all__ = [
    "Workflow",
    "WorkflowValue",
]



from unittest.mock import AsyncMock

import pytest

from workflow_engine import (
    Edge,
    OutputEdge,
    StringValue,
    UserException,
    Workflow,
    WorkflowErrors,
)
from workflow_engine.contexts import InMemoryContext
from workflow_engine.execution import TopologicalExecutionAlgorithm
from workflow_engine.nodes import ConstantStringNode, ErrorNode


@pytest.fixture
def workflow():
    """Helper function to create the error workflow."""
    return Workflow(
        nodes=[
            constant := ConstantStringNode.from_value(id="constant", value="test"),
            error := ErrorNode.from_name(id="error", name="RuntimeError"),
        ],
        edges=[
            Edge.from_nodes(
                source=constant,
                source_key="value",
                target=error,
                target_key="info",
            ),
        ],
        input_edges=[],
        output_edges=[
            OutputEdge.from_node(
                source=constant,
                source_key="value",
                output_key="text",
            ),
        ],
    )


@pytest.mark.unit
def test_workflow_serialization(workflow: Workflow):
    """Test that the error workflow can be serialized and deserialized correctly."""
    workflow_json = workflow.model_dump_json(indent=2)
    with open("examples/error.json", "r") as f:
        assert workflow_json == f.read().strip()

    deserialized_workflow = Workflow.model_validate_json(workflow_json)
    assert deserialized_workflow == workflow


@pytest.mark.asyncio
async def test_workflow_error_handling(workflow: Workflow):
    """Test that the workflow properly handles errors and calls context callbacks."""
    context = InMemoryContext()

    # Create a mock for on_node_error while preserving the original function
    original_on_node_error = context.on_node_error
    mock_on_node_error = AsyncMock(side_effect=original_on_node_error)
    context.on_node_error = mock_on_node_error

    algorithm = TopologicalExecutionAlgorithm()

    errors, output = await algorithm.execute(
        context=context,
        workflow=workflow,
        input={},
    )

    error_node = workflow.nodes[1]
    # Verify the error was captured correctly
    assert errors == WorkflowErrors(
        workflow_errors=[],
        node_errors={error_node.id: ["RuntimeError: test"]},
    )

    # Verify the output still contains the constant value
    assert output == {"text": StringValue("test")}

    # Verify on_node_error was called with the correct arguments
    mock_on_node_error.assert_called_once()
    call_args = mock_on_node_error.call_args
    assert call_args.kwargs["node"] is error_node
    exception = call_args.kwargs["exception"]
    assert isinstance(exception, UserException)
    assert exception.message == "RuntimeError: test"



import pytest

from workflow_engine import (
    BooleanValue,
    Edge,
    InputEdge,
    IntegerValue,
    OutputEdge,
    Workflow,
)
from workflow_engine.contexts import InMemoryContext
from workflow_engine.execution import TopologicalExecutionAlgorithm
from workflow_engine.nodes import AddNode, ConstantIntegerNode, IfElseNode


@pytest.fixture
def add_one_workflow() -> Workflow:
    """Create a workflow that adds one to a number."""

    return Workflow(
        nodes=[
            one := ConstantIntegerNode.from_value(id="one", value=1),
            add_one := AddNode(id="add_one"),
        ],
        edges=[
            Edge.from_nodes(
                source=one,
                source_key="value",
                target=add_one,
                target_key="b",
            ),
        ],
        input_edges=[
            InputEdge.from_node(
                input_key="start",
                target=add_one,
                target_key="a",
            ),
        ],
        output_edges=[
            OutputEdge.from_node(
                source=add_one,
                source_key="sum",
                output_key="result",
            ),
        ],
    )


@pytest.fixture
def subtract_one_workflow() -> Workflow:
    """Create a workflow that subtracts one from a number."""

    return Workflow(
        nodes=[
            negative_one := ConstantIntegerNode.from_value(id="negative_one", value=-1),
            subtract_one := AddNode(id="subtract_one"),
        ],
        edges=[
            Edge.from_nodes(
                source=negative_one,
                source_key="value",
                target=subtract_one,
                target_key="b",
            ),
        ],
        input_edges=[
            InputEdge.from_node(
                input_key="start",
                target=subtract_one,
                target_key="a",
            ),
        ],
        output_edges=[
            OutputEdge.from_node(
                source=subtract_one,
                source_key="sum",
                output_key="result",
            ),
        ],
    )


@pytest.mark.asyncio
async def test_conditional_workflow(
    add_one_workflow: Workflow,
    subtract_one_workflow: Workflow,
):
    """Test that the workflow outputs start+1 when condition is True, and
    start-1 when condition is False."""
    context = InMemoryContext()
    algorithm = TopologicalExecutionAlgorithm()

    start_value = 42

    workflow = Workflow(
        nodes=[
            conditional := IfElseNode.from_workflows(
                id="conditional",
                if_true=add_one_workflow,
                if_false=subtract_one_workflow,
            ),
        ],
        edges=[],
        input_edges=[
            InputEdge.from_node(
                input_key="start",
                target=conditional,
                target_key="start",
            ),
            InputEdge.from_node(
                input_key="condition",
                target=conditional,
                target_key="condition",
            ),
        ],
        output_edges=[
            OutputEdge.from_node(
                source=conditional,
                source_key="result",
                output_key="result",
            ),
        ],
    )

    errors, output = await algorithm.execute(
        context=context,
        workflow=workflow,
        input={
            "start": IntegerValue(start_value),
            "condition": BooleanValue(False),
        },
    )
    assert not errors.any(), errors
    assert output == {"result": start_value - 1}

    errors, output = await algorithm.execute(
        context=context,
        workflow=workflow,
        input={
            "start": IntegerValue(start_value),
            "condition": BooleanValue(True),
        },
    )
    assert not errors.any(), errors
    assert output == {"result": start_value + 1}


@pytest.mark.asyncio
async def test_conditional_workflow_twice_series(
    add_one_workflow: Workflow,
    subtract_one_workflow: Workflow,
):
    """Test that the workflow behaves correctly when condition is called twice
    in series, once with True and once with False."""
    context = InMemoryContext()
    algorithm = TopologicalExecutionAlgorithm()

    start_value = 42

    workflow = Workflow(
        nodes=[
            conditional_1 := IfElseNode.from_workflows(
                id="conditional_1",
                if_true=add_one_workflow,
                if_false=subtract_one_workflow,
            ),
            conditional_2 := IfElseNode.from_workflows(
                id="conditional_2",
                if_true=add_one_workflow,
                if_false=subtract_one_workflow,
            ),
        ],
        edges=[
            Edge.from_nodes(
                source=conditional_1,
                source_key="result",
                target=conditional_2,
                target_key="start",
            ),
        ],
        input_edges=[
            InputEdge.from_node(
                input_key="start",
                target=conditional_1,
                target_key="start",
            ),
            InputEdge.from_node(
                input_key="condition_1",
                target=conditional_1,
                target_key="condition",
            ),
            InputEdge.from_node(
                input_key="condition_2",
                target=conditional_2,
                target_key="condition",
            ),
        ],
        output_edges=[
            OutputEdge.from_node(
                source=conditional_2,
                source_key="result",
                output_key="result",
            ),
        ],
    )

    errors, output = await algorithm.execute(
        context=context,
        workflow=workflow,
        input={
            "start": IntegerValue(start_value),
            "condition_1": BooleanValue(True),
            "condition_2": BooleanValue(False),
        },
    )
    assert not errors.any(), errors
    assert output == {"result": start_value}

    errors, output = await algorithm.execute(
        context=context,
        workflow=workflow,
        input={
            "start": IntegerValue(start_value),
            "condition_1": BooleanValue(False),
            "condition_2": BooleanValue(True),
        },
    )
    assert not errors.any(), errors
    assert output == {"result": start_value}

    errors, output = await algorithm.execute(
        context=context,
        workflow=workflow,
        input={
            "start": IntegerValue(start_value),
            "condition_1": BooleanValue(True),
            "condition_2": BooleanValue(True),
        },
    )
    assert not errors.any(), errors
    assert output == {"result": start_value + 2}

    errors, output = await algorithm.execute(
        context=context,
        workflow=workflow,
        input={
            "start": IntegerValue(start_value),
            "condition_1": BooleanValue(False),
            "condition_2": BooleanValue(False),
        },
    )
    assert not errors.any(), errors
    assert output == {"result": start_value - 2}



# tests/test_schema.py

"""
Tests two sets of functionalities:
1. that .to_value_cls() is the inverse of .to_value_schema()
2. that we can manually write JSON Schemas that will turn into the correct Value
   classes when .to_value_cls() is called on them.
3. that the aliasing system (e.g. { "title": "StringValue" } -> StringValue)
   works for all non-generic Value classes.

In this file we follow the convention of using T for the expected type and U for
the type returned by .to_value_cls().
"""

import pytest

from workflow_engine import (
    BooleanValue,
    Data,
    Empty,
    FileValue,
    FloatValue,
    IntegerValue,
    NullValue,
    SequenceValue,
    StringMapValue,
    StringValue,
    ValueSchemaValue,
    WorkflowValue,
)
from workflow_engine.core.values import validate_value_schema
from workflow_engine.core.values.schema import (
    BooleanValueSchema,
    BaseValueSchema,
    FloatValueSchema,
    IntegerValueSchema,
    NullValueSchema,
    ReferenceValueSchema,
    SequenceValueSchema,
    StringMapValueSchema,
    StringValueSchema,
)
from workflow_engine.files import (
    JSONFileValue,
    JSONLinesFileValue,
    PDFFileValue,
    TextFileValue,
)

# ensure that these node types are registered for the workflow tests
from workflow_engine.nodes import (
    AddNode,  # noqa: F401
    ConstantIntegerNode,  # noqa: F401
)


@pytest.mark.unit
def test_boolean_schema_roundtrip():
    T = BooleanValue
    schema = T.to_value_schema()
    assert isinstance(schema, BooleanValueSchema)
    U = schema.to_value_cls()
    assert U == T
    assert U.to_value_schema() == schema

    t1 = T(True)
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U(False)
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_boolean_schema_manual():
    T = BooleanValue
    json_schema = {
        "type": "boolean",
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, BooleanValueSchema)
    U = schema.to_value_cls()
    assert U == T

    t1 = T(False)
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U(True)
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_boolean_schema_aliasing():
    T = BooleanValue
    json_schema = {
        "title": "BooleanValue",
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, BaseValueSchema)
    U = schema.to_value_cls()
    assert U == BooleanValue

    t1 = T(True)
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U(False)
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_integer_schema_roundtrip():
    T = IntegerValue
    schema = T.to_value_schema()
    assert isinstance(schema, IntegerValueSchema)
    U = schema.to_value_cls()
    assert U == T
    assert U.to_value_schema() == schema

    t1 = T(42)
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U(2520)
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_integer_schema_manual():
    T = IntegerValue
    json_schema = {
        "type": "integer",
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, IntegerValueSchema)
    U = schema.to_value_cls()
    assert U == T

    t1 = T(-42)
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U(2520)
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_integer_schema_aliasing():
    T = IntegerValue
    json_schema = {
        "title": "IntegerValue",
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, BaseValueSchema)
    U = schema.to_value_cls()
    assert U == T

    t1 = T(2048)
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U(2520)
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_float_schema_roundtrip():
    T = FloatValue
    schema = T.to_value_schema()
    assert isinstance(schema, FloatValueSchema)
    U = schema.to_value_cls()
    assert U == T
    assert U.to_value_schema() == schema

    t1 = T(3.14159)
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U(2.71828)
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_float_schema_manual():
    T = FloatValue
    json_schema = {
        "type": "number",
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, FloatValueSchema)
    U = schema.to_value_cls()
    assert U == T

    t1 = T(2.71828)
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U(3.14159)
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_float_schema_aliasing():
    T = FloatValue
    json_schema = {
        "title": "FloatValue",
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, BaseValueSchema)
    U = schema.to_value_cls()
    assert U == FloatValue

    t1 = T(1.41421)
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U(3.14159)
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_null_schema_roundtrip():
    T = NullValue
    schema = T.to_value_schema()
    assert isinstance(schema, NullValueSchema)
    U = schema.to_value_cls()
    assert U == T
    assert U.to_value_schema() == schema

    t1 = T(None)
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U(None)
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_null_schema_manual():
    T = NullValue
    json_schema = {
        "type": "null",
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, NullValueSchema)
    U = schema.to_value_cls()
    assert U == T

    t1 = T(None)
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U(None)
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_null_schema_aliasing():
    T = NullValue
    json_schema = {
        "title": "NullValue",
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, BaseValueSchema)
    U = schema.to_value_cls()
    assert U == T

    t1 = T(None)
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U(None)
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_string_schema_roundtrip():
    T = StringValue
    schema = T.to_value_schema()
    assert isinstance(schema, StringValueSchema)
    U = schema.to_value_cls()
    assert U == StringValue
    assert U.to_value_schema() == schema

    t1 = T("hello wengine")
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U("hi wengine")
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_string_schema_manual():
    T = StringValue
    json_schema = {
        "type": "string",
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, StringValueSchema)
    U = schema.to_value_cls()
    assert U == T

    t1 = T("salutations wengine")
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U("good morning wengine")
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_string_schema_aliasing():
    T = StringValue
    json_schema = {
        "title": "StringValue",
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, BaseValueSchema)
    U = schema.to_value_cls()
    assert U == StringValue

    t1 = T("hey wengine")
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U("good afternoon wengine")
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_sequence_schema_roundtrip():
    for ItemType, ItemSchema in (
        (BooleanValue, BooleanValueSchema),
        (FloatValue, FloatValueSchema),
        (IntegerValue, IntegerValueSchema),
        (NullValue, NullValueSchema),
        (StringValue, StringValueSchema),
    ):
        T = SequenceValue[ItemType]
        schema = T.to_value_schema()
        assert isinstance(schema, SequenceValueSchema)
        assert isinstance(schema.items, ReferenceValueSchema)
        assert isinstance(schema.defs[schema.items.id], ItemSchema)
        U = schema.to_value_cls()
        assert U == T
        assert U.to_value_schema() == schema


@pytest.mark.unit
def test_sequence_schema_manual():
    for type, ItemType, ItemSchema in (
        ("boolean", BooleanValue, BooleanValueSchema),
        ("number", FloatValue, FloatValueSchema),
        ("integer", IntegerValue, IntegerValueSchema),
        ("null", NullValue, NullValueSchema),
        ("string", StringValue, StringValueSchema),
    ):
        T = SequenceValue[ItemType]
        json_schema = {
            "type": "array",
            "items": {"type": type},
        }
        schema = validate_value_schema(json_schema)
        assert isinstance(schema, SequenceValueSchema)
        assert isinstance(schema.items, ItemSchema)
        U = schema.to_value_cls()
        assert U == T


@pytest.mark.unit
def test_sequence_schema_aliasing():
    for ItemType in (
        BooleanValue,
        FloatValue,
        IntegerValue,
        NullValue,
        StringValue,
    ):
        T = SequenceValue[ItemType]
        json_schema = {
            "type": "array",
            "items": {"title": ItemType.__name__},
        }
        schema = validate_value_schema(json_schema)
        assert isinstance(schema, SequenceValueSchema)
        assert isinstance(schema.items, BaseValueSchema)
        U = schema.to_value_cls()
        assert U == T


@pytest.mark.unit
def test_string_map_schema_roundtrip():
    for ItemType, ItemSchema in (
        (BooleanValue, BooleanValueSchema),
        (FloatValue, FloatValueSchema),
        (IntegerValue, IntegerValueSchema),
        (NullValue, NullValueSchema),
        (StringValue, StringValueSchema),
    ):
        T = StringMapValue[ItemType]
        schema = T.to_value_schema()
        assert isinstance(schema, StringMapValueSchema)
        assert isinstance(schema.additionalProperties, ReferenceValueSchema)
        assert isinstance(schema.defs[schema.additionalProperties.id], ItemSchema)
        U = schema.to_value_cls()
        assert U == T
        assert U.to_value_schema() == schema


@pytest.mark.unit
def test_string_map_schema_manual():
    for type, ItemType, ItemSchema in (
        ("boolean", BooleanValue, BooleanValueSchema),
        ("number", FloatValue, FloatValueSchema),
        ("integer", IntegerValue, IntegerValueSchema),
        ("null", NullValue, NullValueSchema),
        ("string", StringValue, StringValueSchema),
    ):
        T = StringMapValue[ItemType]
        json_schema = {
            "type": "object",
            "additionalProperties": {"type": type},
        }
        schema = validate_value_schema(json_schema)
        assert isinstance(schema, StringMapValueSchema)
        assert isinstance(schema.additionalProperties, ItemSchema)
        U = schema.to_value_cls()
        assert U == T


@pytest.mark.unit
def test_string_map_schema_aliasing():
    for ItemType in (
        BooleanValue,
        FloatValue,
        IntegerValue,
        NullValue,
        StringValue,
    ):
        T = StringMapValue[ItemType]
        json_schema = {
            "type": "object",
            "additionalProperties": {"title": ItemType.__name__},
        }
        schema = validate_value_schema(json_schema)
        assert isinstance(schema, StringMapValueSchema)
        assert isinstance(schema.additionalProperties, BaseValueSchema)
        U = schema.to_value_cls()
        assert U == T


@pytest.mark.unit
def test_super_recursive_schema_roundtrip():
    for T in (
        StringMapValue[SequenceValue[StringMapValue[StringValue]]],
        SequenceValue[StringMapValue[SequenceValue[NullValue]]],
        StringMapValue[StringMapValue[StringMapValue[IntegerValue]]],
        SequenceValue[SequenceValue[SequenceValue[BooleanValue]]],
    ):
        schema = T.to_value_schema()
        U = schema.to_value_cls()
        assert U == T
        assert U.to_value_schema() == schema


@pytest.mark.unit
def test_super_recursive_schema_manual():
    T = StringMapValue[SequenceValue[StringMapValue[StringValue]]]
    json_schema = {
        "type": "object",
        "additionalProperties": {
            "type": "array",
            "items": {
                "type": "object",
                "additionalProperties": {"title": "StringValue"},
            },
        },
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, StringMapValueSchema)
    assert isinstance(schema.additionalProperties, SequenceValueSchema)
    assert isinstance(schema.additionalProperties.items, StringMapValueSchema)
    assert isinstance(
        schema.additionalProperties.items.additionalProperties, BaseValueSchema
    )
    U = schema.to_value_cls()
    assert U == T


@pytest.mark.unit
def test_empty_schema_roundtrip():
    T = Empty
    schema = T.to_value_schema()
    U = schema.to_value_cls()

    # for Empty, to_value_cls returns a new class not equal to the original
    # but it can serialize and deserialize instances of the original class
    t1 = T()
    u1 = U.model_validate(t1.model_dump())
    # equality check fails because they are technically different classes,
    # but they have the exact same fields
    assert u1.root.__dict__ == t1.__dict__

    u2 = U.model_validate({})
    t2 = T.model_validate(u2.model_dump())
    # equality check fails because they are technically different classes,
    # but they have the exact same fields
    assert t2.__dict__ == u2.root.__dict__


# defined outside of test_data_schema_roundtrip to get a proper class name
class FooBarData(Data):
    foo: StringValue
    bar: IntegerValue


@pytest.mark.unit
def test_data_schema_roundtrip():
    T = FooBarData
    schema = T.to_value_schema()
    U = schema.to_value_cls()

    # it can serialize and deserialize instances of the original class
    t1 = T(
        foo=StringValue("foo"),
        bar=IntegerValue(1),
    )
    u1 = U.model_validate(t1.model_dump())
    # equality check fails because they are technically different classes,
    # but they have the exact same fields
    assert u1.root.foo == t1.foo
    assert u1.root.bar == t1.bar
    assert u1.root.__dict__ == t1.__dict__

    u2 = U.model_validate(
        {
            "foo": "bar",
            "bar": 2,
        }
    )
    t2 = T.model_validate(u2.model_dump())
    # equality check fails because they are technically different classes,
    # but they have the exact same fields
    assert t2.foo == u2.root.foo
    assert t2.bar == u2.root.bar


@pytest.mark.unit
def test_data_schema_manual():
    T = FooBarData
    json_schema = {
        "title": "FooBarData",
        "type": "object",
        "properties": {
            "foo": {"type": "string"},
            "bar": {"type": "integer"},
        },
        "required": ["foo", "bar"],
    }
    schema = validate_value_schema(json_schema)
    U = schema.to_value_cls()

    # it can serialize and deserialize instances of the original class
    t1 = T(
        foo=StringValue("bar"),
        bar=IntegerValue(12),
    )
    u1 = U.model_validate(t1.model_dump())
    # equality check fails because they are technically different classes,
    # but they have the exact same fields
    assert u1.root.foo == t1.foo
    assert u1.root.bar == t1.bar

    u2 = U.model_validate(
        {
            "foo": "baz",
            "bar": 24,
        }
    )
    t2 = T.model_validate(u2.model_dump())
    # equality check fails because they are technically different classes,
    # but they have the exact same fields
    assert u2.root.foo == t2.foo
    assert u2.root.bar == t2.bar


@pytest.mark.unit
def test_data_schema_aliasing():
    T = FooBarData
    json_schema = {
        "title": "FooBarData",
        "type": "object",
        "properties": {
            "foo": {"title": "StringValue"},
            "bar": {"title": "IntegerValue"},
        },
        "required": ["foo", "bar"],
    }
    schema = validate_value_schema(json_schema)
    U = schema.to_value_cls()

    # it can serialize and deserialize instances of the original class
    t1 = T(
        foo=StringValue("foobar"),
        bar=IntegerValue(123),
    )
    u1 = U.model_validate(t1.model_dump())
    # equality check fails because they are technically different classes,
    # but they have the exact same fields
    assert u1.root.foo == t1.foo
    assert u1.root.bar == t1.bar

    u2 = U.model_validate(
        {
            "foo": "baz",
            "bar": 24,
        }
    )
    t2 = T.model_validate(u2.model_dump())
    assert u2.root.foo == t2.foo
    assert u2.root.bar == t2.bar


@pytest.mark.unit
def test_file_schema_roundtrip():
    for T in (
        FileValue,
        JSONFileValue,
        JSONLinesFileValue,
        PDFFileValue,
        TextFileValue,
    ):
        schema = T.to_value_schema()
        U = schema.to_value_cls()
        assert U == T
        assert U.to_value_schema() == schema

        t1 = T.from_path("foo", foo="bar", bar="baz")
        u1 = U.model_validate(t1.model_dump())
        assert u1 == t1

        u2 = U.model_validate({"path": "bar", "metadata": {"baz": 3}})
        t2 = T.model_validate(u2.model_dump())
        assert t2.path == u2.root.path
        assert t2.metadata == u2.root.metadata


@pytest.mark.unit
def test_file_schema_aliasing():
    for T in (
        FileValue,
        JSONFileValue,
        JSONLinesFileValue,
        PDFFileValue,
        TextFileValue,
    ):
        json_schema = {"title": T.__name__}
        schema = validate_value_schema(json_schema)
        U = schema.to_value_cls()
        assert U == T

        t1 = T.from_path("bar", bar="baz", baz="foo")
        u1 = U.model_validate(t1.model_dump())
        assert u1.root.path == t1.path
        assert u1.root.metadata == t1.metadata

        u2 = U.model_validate({"path": "bar", "metadata": {"baz": 3}})
        t2 = T.model_validate(u2.model_dump())
        assert t2.path == u2.root.path
        assert t2.metadata == u2.root.metadata


@pytest.mark.unit
def test_workflow_schema_roundtrip():
    T = WorkflowValue
    schema = T.to_value_schema()
    U = schema.to_value_cls()
    assert U == T
    assert U.to_value_schema() == schema

    with open("examples/addition.json", "r") as f:
        workflow_json = f.read().strip()
    t1 = T.model_validate_json(workflow_json)
    u1 = U.model_validate_json(workflow_json)
    assert u1 == t1


@pytest.mark.unit
def test_workflow_schema_aliasing():
    T = WorkflowValue
    json_schema = {"title": T.__name__}
    schema = validate_value_schema(json_schema)
    U = schema.to_value_cls()
    assert U == T

    with open("examples/addition.json", "r") as f:
        workflow_json = f.read().strip()

    t1 = T.model_validate_json(workflow_json)
    u1 = U.model_validate_json(workflow_json)
    assert u1 == t1


@pytest.mark.unit
def test_value_schema_value_roundtrip():
    T = ValueSchemaValue
    schema = T.to_value_schema()
    U = schema.to_value_cls()
    assert U == T
    assert U.to_value_schema() == schema

    t1 = T(StringValueSchema(type="string", title="StringValue"))
    u1 = U.model_validate(t1.model_dump())
    assert u1 == t1

    u2 = U(IntegerValueSchema(type="integer", title="IntegerValue"))
    t2 = T.model_validate(u2.model_dump())
    assert t2 == u2


@pytest.mark.unit
def test_value_schema_value_manual():
    T = ValueSchemaValue
    json_schema = {"title": T.__name__}
    schema = validate_value_schema(json_schema)
    U = schema.to_value_cls()
    assert U == T

    t1 = T(BooleanValueSchema(type="boolean", title="BooleanValue"))
    u1 = U.model_validate(t1.model_dump())
    assert u1 == t1

    u2 = U(NullValueSchema(type="null", title="NullValue"))
    t2 = T.model_validate(u2.model_dump())
    assert t2 == u2


@pytest.mark.unit
def test_value_schema_value_aliasing():
    T = ValueSchemaValue
    json_schema = {"title": T.__name__}
    schema = validate_value_schema(json_schema)
    U = schema.to_value_cls()
    assert U == T

    t1 = T(FloatValueSchema(type="number", title="FloatValue"))
    u1 = U.model_validate(t1.model_dump())
    assert u1 == t1

    u2 = U(
        SequenceValueSchema(
            type="array",
            items=StringValueSchema(type="string", title="StringValue"),
        )
    )
    t2 = T.model_validate(u2.model_dump())
    assert t2 == u2



# workflow_engine/core/values/value.py
import inspect
import re
from functools import cached_property
from hashlib import md5
from logging import getLogger
from typing import (
    TYPE_CHECKING,
    Awaitable,
    ClassVar,
    Generic,
    Protocol,
    Self,
    Type,
    TypeVar,
    get_origin,
)

from pydantic import PrivateAttr

from ...utils.immutable import ImmutableRootModel

if TYPE_CHECKING:
    from ..context import Context
    from .schema import ValueSchema


logger = getLogger(__name__)


T = TypeVar("T")
V = TypeVar("V", bound="Value")
type ValueType = Type[Value]


def get_origin_and_args(t: ValueType) -> tuple[ValueType, tuple[ValueType, ...]]:
    """
    For a non-generic value type NonGenericValue, returns (NonGenericValue, ()).

    For a generic value type GenericValue[Argument1Value, Argument2Value, ...],
    returns (GenericValue, (Argument1Value, Argument2Value, ...)).
    All arguments must themselves be Value subclasses.
    """
    # Pydantic RootModels don't play nice with get_origin and get_args, so we
    # get the root type directly from the model fields.
    assert issubclass(t, Value)
    info = t.__pydantic_generic_metadata__
    origin = info["origin"]
    args = info["args"]
    if origin is None:
        assert len(args) == 0
        return t, ()
    else:
        assert issubclass(origin, Value)
        assert len(args) > 0
        return origin, tuple(args)


type ValueTypeKey = tuple[str, tuple[ValueTypeKey, ...]]


def get_value_type_key(t: ValueType) -> ValueTypeKey:
    """
    Get a unique hashable key for a Value type.
    If t is a generic type, recursively call `get_value_type_key` to expand any
    Value types in the args.
    """
    origin, args = get_origin_and_args(t)
    return origin.__name__, tuple(
        get_value_type_key(arg) if issubclass(arg, Value) else arg for arg in args
    )


SourceType = TypeVar("SourceType", bound="Value")
TargetType = TypeVar("TargetType", bound="Value")


class Caster(Protocol, Generic[SourceType, TargetType]):  # type: ignore
    """
    A caster is a contextual function that transforms the type of a Value.

    May be either a sync or async function.
    """

    def __call__(
        self,
        value: SourceType,
        context: "Context",
    ) -> TargetType | Awaitable[TargetType]: ...


class GenericCaster(Protocol, Generic[SourceType, TargetType]):  # type: ignore
    """
    A generic caster is a contextual function that takes a source type and a
    target type, and outputs a caster between the two types, or None if the cast
    is not possible.
    This is an advanced feature intended for use on generic types.

    The purpose of this two-step approach is to explicitly allow or deny type
    casts before the exact type of the value is known. This is necessary because
    the type of a Value is not known until the Value is created.
    """

    def __call__(
        self,
        source_type: Type[SourceType],
        target_type: Type[TargetType],
    ) -> Caster[SourceType, TargetType] | None: ...


generic_pattern = re.compile(r"^[a-zA-Z]\w+\[.*\]$")


class Value(ImmutableRootModel[T], Generic[T]):
    """
    Wraps an arbitrary read-only value which can be passed as input to a node.

    Each Value subclass defines a specific type (possibly generic) of value.
    After defining the subclass, you can register Caster functions to convert
    other Value classes to that type, using the register_cast_to decorator.
    Casts are registered in any order.

    Each Value subclass inherits its parent classes' Casters.
    To avoid expanding the type tree every time, we cache the Casters at each
    class the first time a cast is used.
    Once that cache is created, the casts are locked and can no longer be
    changed.
    """

    # these properties force us to implement __eq__ and __hash__ to ignore them
    _casters: ClassVar[dict[str, GenericCaster]] = {}
    _resolved_casters: ClassVar[dict[str, GenericCaster] | None] = None
    _cast_cache: dict[ValueTypeKey, "Value"] = PrivateAttr(
        default_factory=dict,
    )

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        # reinitialize for each subclass so it doesn't just reference the parent
        cls._casters = {}
        cls._resolved_casters = None

        # NOTE: something about this hack does not work when using
        # `from __future__ import annotations`.
        while generic_pattern.match(cls.__name__) is not None:
            assert cls.__base__ is not None
            cls = cls.__base__
        if get_origin(cls) is None:
            value_type_registry.register(cls.__name__, cls)

    @classmethod
    def _get_casters(cls) -> dict[str, GenericCaster]:
        """
        Get all type casting functions for this class, including those inherited
        from parent classes.
        This inherits from all parents classes, though they will be overridden
        if the child class has its own casting function for the same type.
        """
        if cls._resolved_casters is not None:
            return cls._resolved_casters

        resolved_casters: dict[str, GenericCaster] = cls._casters.copy()

        # Add converters from all classes in MRO order
        # (starting from this class, then parents, then parents of parents, ...)
        for parent in cls.__bases__:
            if issubclass(parent, Value):
                parent_casters = parent._get_casters()
            else:
                continue

            for origin, caster in parent_casters.items():
                # converters in the child class override those in the parent
                # class
                if origin not in resolved_casters:
                    resolved_casters[origin] = caster

        cls._resolved_casters = resolved_casters
        return resolved_casters

    @classmethod
    def register_cast_to(cls, t: Type[V]):
        """
        A decorator to register a possible type cast from this class to the
        class T, neither of which are generic.
        """

        def wrap(caster: Caster[Self, V]):
            cls.register_generic_cast_to(t)(lambda source_type, target_type: caster)
            return caster

        return wrap

    @classmethod
    def register_generic_cast_to(cls, t: Type[V]):
        """
        A decorator to register a possible type cast from this class to the
        class T, either of which may be generic.
        """

        def wrap(caster: GenericCaster[Self, V]):
            if cls._resolved_casters is not None:
                raise RuntimeError(
                    f"Cannot add casters for {cls.__name__} after it has been used to cast values"
                )

            target_origin, _ = get_origin_and_args(t)
            name = target_origin.__name__
            if name in cls._casters:
                raise AssertionError(
                    f"Type caster from {cls.__name__} to {name} already registered"
                )
            cls._casters[name] = caster

        return wrap

    @classmethod
    def get_caster(cls, t: Type[V]) -> Caster[Self, V] | None:
        converters = cls._get_casters()
        target_origin, _ = get_origin_and_args(t)
        if target_origin.__name__ in converters:
            generic_caster = converters[target_origin.__name__]
            caster = generic_caster(cls, t)
            if caster is not None:
                return caster

        if issubclass(cls, t):
            return lambda value, context: value  # type: ignore

        return None

    @classmethod
    def can_cast_to(cls, t: Type[V]) -> bool:
        """
        Returns True if there is any hope of casting this value to the type t.
        """
        return cls.get_caster(t) is not None

    def __eq__(self, other) -> bool:
        if isinstance(other, Value):
            return self.root == other.root
        return self.root == other

    def __hash__(self):
        return hash(self.root)

    def __str__(self) -> str:
        return str(self.root)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.root})"

    def __bool__(self) -> bool:
        return bool(self.root)

    @cached_property
    def md5(self) -> str:
        return md5(str(self).encode()).hexdigest()

    async def cast_to(self, t: Type[V], *, context: "Context") -> V:
        key = get_value_type_key(t)
        if key in self._cast_cache:
            casted: V = self._cast_cache[key]  # type: ignore
            return casted

        cast_fn = self.__class__.get_caster(t)
        if cast_fn is not None:
            result = cast_fn(self, context)
            casted: V = (await result) if inspect.iscoroutine(result) else result  # type: ignore
            self._cast_cache[key] = casted
            return casted

        raise ValueError(f"Cannot convert {self} to {t}")

    @classmethod
    async def cast_from(cls, v: "Value", *, context: "Context") -> Self:
        return await v.cast_to(cls, context=context)

    @classmethod
    def to_value_schema(cls) -> "ValueSchema":
        from .schema import validate_value_schema  # avoid circular import

        return validate_value_schema(cls.model_json_schema())


class ValueRegistry:
    def __init__(self):
        self.types: dict[str, ValueType] = {}

    def register(self, name: str, cls: ValueType):
        if name in self.types:
            conflict = self.types[name]
            if cls is not conflict:
                raise ValueError(
                    f'Value type "{name}" (class {cls.__name__}) is already registered to a different class ({conflict.__name__})'
                )
        self.types[name] = cls
        logger.debug("Registering class %s as value type %s", cls.__name__, name)

    def __contains__(self, name: str) -> bool:
        return name in self.types

    def __getitem__(self, name: str) -> ValueType:
        if name not in self.types:
            raise ValueError(f'Value type "{name}" is not registered')
        return self.types[name]


value_type_registry = ValueRegistry()


__all__ = [
    "Caster",
    "GenericCaster",
    "get_origin_and_args",
    "Value",
    "ValueType",
    "value_type_registry",
]



# workflow_engine/nodes/iteration.py
"""
Nodes that iterate over a sequence of items.
"""

from typing import ClassVar, Literal, Self, Type

from overrides import override

from ..core import (
    Context,
    DataValue,
    Edge,
    InputEdge,
    Node,
    NodeTypeInfo,
    OutputEdge,
    Params,
    Workflow,
    WorkflowValue,
)
from .data import (
    ExpandDataNode,
    ExpandSequenceNode,
    GatherDataNode,
    GatherSequenceNode,
    SequenceData,
)


class ForEachParams(Params):
    workflow: WorkflowValue


class ForEachNode(Node[SequenceData, SequenceData, ForEachParams]):
    """
    A node that executes the internal workflow W for each item in the input
    sequence.

    For each item i in the input sequence, create a copy of W, call it W[i].
    We expand the sequence into its individual data objects and expand
    sequence[i] into the input fields of W[i].
    Then, we gather the output of each W[i] into a single object, before
    gathering them further into a single sequence.

    The output of this node is a sequence of the same length as the input
    sequence, with each item being the output of the internal workflow.
    """

    TYPE_INFO: ClassVar[NodeTypeInfo] = NodeTypeInfo.from_parameter_type(
        name="ForEach",
        display_name="ForEach",
        description="Executes the internal workflow for each item in the input sequence.",
        version="0.4.0",
        parameter_type=ForEachParams,
    )

    type: Literal["ForEach"] = "ForEach"  # pyright: ignore[reportIncompatibleVariableOverride]

    @property
    def workflow(self) -> Workflow:
        return self.params.workflow.root

    @property
    @override
    def input_type(self) -> Type[SequenceData]:
        return SequenceData[DataValue[self.workflow.input_type]]

    @property
    @override
    def output_type(self) -> Type[SequenceData]:
        return SequenceData[DataValue[self.workflow.output_type]]

    @override
    async def run(self, context: Context, input: SequenceData) -> Workflow:
        N = len(input.sequence)

        nodes: list[Node] = []
        edges: list[Edge] = []

        expand = ExpandSequenceNode.from_length(
            id="expand",
            length=N,
            element_type=DataValue[self.workflow.input_type],
        )
        gather = GatherSequenceNode.from_length(
            id="gather",
            length=N,
            element_type=DataValue[self.workflow.output_type],
        )
        nodes.append(expand)
        nodes.append(gather)

        for i in range(N):
            namespace = f"element_{i}"
            input_adapter = ExpandDataNode.from_data_type(
                id="input_adapter",
                data_type=self.workflow.input_type,
            ).with_namespace(namespace)
            item_workflow = self.workflow.with_namespace(namespace)
            output_adapter = GatherDataNode.from_data_type(
                id="output_adapter",
                data_type=self.workflow.output_type,
            ).with_namespace(namespace)

            nodes.append(input_adapter)
            nodes.extend(item_workflow.nodes)
            nodes.append(output_adapter)

            edges.append(
                Edge.from_nodes(
                    source=expand,
                    source_key=expand.key(i),
                    target=input_adapter,
                    target_key="data",
                )
            )
            for input_edge in item_workflow.input_edges:
                edges.append(
                    Edge(
                        source_id=input_adapter.id,
                        source_key=input_edge.input_key,
                        target_id=input_edge.target_id,
                        target_key=input_edge.target_key,
                    )
                )
            edges.extend(item_workflow.edges)
            for output_edge in item_workflow.output_edges:
                edges.append(
                    Edge(
                        source_id=output_edge.source_id,
                        source_key=output_edge.source_key,
                        target_id=output_adapter.id,
                        target_key=output_edge.output_key,
                    )
                )
            edges.append(
                Edge.from_nodes(
                    source=output_adapter,
                    source_key="data",
                    target=gather,
                    target_key=gather.key(i),
                )
            )

        return Workflow(
            nodes=nodes,
            edges=edges,
            input_edges=[
                InputEdge.from_node(
                    input_key="sequence",
                    target=expand,
                    target_key="sequence",
                )
            ],
            output_edges=[
                OutputEdge.from_node(
                    source=gather,
                    source_key="sequence",
                    output_key="sequence",
                )
            ],
        )

    @classmethod
    def from_workflow(
        cls,
        id: str,
        workflow: Workflow,
    ) -> Self:
        return cls(id=id, params=ForEachParams(workflow=WorkflowValue(workflow)))


__all__ = [
    "ForEachNode",
]
